""" Test the CLI for run mip-dna """
import logging

from cg.cli.workflow.mip_dna.base import run
from cg.apps.mip import MipAPI

CASE_ID = "yellowhog"
CONDA_ENV = "S_mip_rd-dna"
EMAIL = "james.holden@scilifelab.se"


def test_cg_dry_run(cli_runner, tb_api, analysis_store_single_case, mip_context, caplog):
    """Test print the MIP run to console"""
    # GIVEN a cli function
    with caplog.at_level(logging.INFO):
        # WHEN we run a case in dry run mode
        result = cli_runner.invoke(run, ["--dry-run", "--email", EMAIL, CASE_ID], obj=mip_context)
        # THEN command is run successfully
        assert result.exit_code == 0
        # THEN the command should be printed
        assert (
            "analyse rd_dna yellowhog --config config.yaml "
            "--slurm_quality_of_service normal --email james.holden@scilifelab.se" in caplog.text
        )



""" Add CLI support to start MIP rare disease RNA"""

import logging

import click

from cg.apps import hk, lims, scoutapi, tb
from cg.apps.environ import environ_email
from cg.cli.workflow.get_links import get_links
from cg.cli.workflow.mip.store import store as store_cmd
from cg.cli.workflow.mip_rna.deliver import CASE_TAGS, SAMPLE_TAGS
from cg.cli.workflow.mip_rna.deliver import deliver as deliver_cmd
from cg.meta.deliver import DeliverAPI
from cg.meta.workflow.mip import AnalysisAPI
from cg.store import Store
from cg.store.utils import case_exists

LOG = logging.getLogger(__name__)


@click.group("mip-rna")
@click.pass_context
def mip_rna(context: click.Context):
    """Rare disease RNA workflow"""

    context.obj["rna_api"] = AnalysisAPI(
        db=Store(context.obj["database"]),
        hk_api=hk.HousekeeperAPI(context.obj),
        tb_api=tb.TrailblazerAPI(context.obj),
        scout_api=scoutapi.ScoutAPI(context.obj),
        lims_api=lims.LimsAPI(context.obj),
        deliver_api=DeliverAPI(
            context.obj,
            hk_api=hk.HousekeeperAPI(context.obj),
            lims_api=lims.LimsAPI(context.obj),
            case_tags=CASE_TAGS,
            sample_tags=SAMPLE_TAGS,
        ),
        script=context.obj["mip-rd-rna"]["script"],
        pipeline=context.obj["mip-rd-rna"]["pipeline"],
        conda_env=context.obj["mip-rd-rna"]["conda_env"],
        root=context.obj["mip-rd-rna"]["root"],
    )


@mip_rna.command()
@click.option("-c", "--case", "case_id", help="link all samples for a case")
@click.argument("sample_id", required=False)
@click.pass_context
def link(context: click.Context, case_id: str, sample_id: str):
    """Link FASTQ files for a sample_id"""
    rna_api = context.obj["rna_api"]
    link_objs = get_links(rna_api.db, case_id, sample_id)

    for link_obj in link_objs:
        LOG.info(
            "%s: %s link FASTQ files", link_obj.sample.internal_id, link_obj.sample.data_analysis
        )
        if "mip + rna" in link_obj.sample.data_analysis.lower():
            rna_api.link_sample(
                sample=link_obj.sample,
                case_id=case_id,
            )


@mip_rna.command()
@click.option("-d", "--dry-run", "dry_run", is_flag=True, help="print command to console")
@click.option("-e", "--email", help="email to send errors to")
@click.option("--mip-dry-run", "mip_dry_run", is_flag=True, help="Run MIP in dry-run mode")
@click.option("-p", "--priority", type=click.Choice(["low", "normal", "high"]))
@click.option("-sw", "--start-with", help="start mip from this program.")
@click.argument("case_id")
@click.pass_context
def run(
    context: click.Context,
    case_id: str,
    dry_run: bool = False,
    mip_dry_run: bool = False,
    priority: str = None,
    email: str = None,
    start_with: str = None,
):
    """Run the analysis for a case"""
    rna_api = context.obj["rna_api"]
    case_obj = rna_api.db.family(case_id)

    if not case_exists(case_obj, case_id):
        raise click.Abort()
    if rna_api.tb.analyses(family=case_obj.internal_id, temp=True).first():
        LOG.warning("%s: analysis already running", case_obj.internal_id)
        return

    email = email or environ_email()
    kwargs = dict(
        config=context.obj["mip-rd-rna"]["mip_config"],
        case=case_id,
        priority=priority,
        email=email,
        dryrun=mip_dry_run,
        start_with=start_with,
    )
    if dry_run:
        rna_api.run_command(dry_run=dry_run, **kwargs)
        return

    rna_api.run_command(**kwargs)

    if mip_dry_run:
        LOG.info("Executed MIP in dry-run mode - skipping Trailblazer step")
        return

    rna_api.tb.mark_analyses_deleted(case_id=case_id)
    rna_api.tb.add_pending_analysis(case_id=case_id, email=email)
    rna_api.set_statusdb_action(case_id=case_id, action="running")
    LOG.info("MIP rd-rna run started!")


@mip_rna.command("config-case")
@click.option("-d", "--dry", is_flag=True, help="Print config to console")
@click.argument("case_id")
@click.pass_context
def config_case(context: click.Context, case_id: str, dry: bool = False):
    """Generate a config for the case_id"""
    rna_api = context.obj["rna_api"]

    case_obj = rna_api.db.family(case_id)
    if not case_exists(case_obj, case_id):
        context.abort()
    config_data = rna_api.pedigree_config(case_obj, pipeline="mip-rna")
    if dry:
        print(config_data)
        return
    out_path = rna_api.write_pedigree_config(config_data)
    LOG.info(f"Config saved to: {out_path}")


mip_rna.add_command(store_cmd)
mip_rna.add_command(deliver_cmd)



""" Trailblazer API for cg """ ""

import logging
from trailblazer.store import Store

LOG = logging.getLogger(__name__)


class TrailblazerAPI(Store):
    """Interface to Trailblazer for `cg`."""

    def __init__(self, config: dict):
        super(TrailblazerAPI, self).__init__(
            config["trailblazer"]["database"],
            families_dir=config["trailblazer"]["root"],
        )



"""
    API for uploading observations
"""

from pathlib import Path
import logging
from typing import List

from cg.apps import hk, loqus
from cg.exc import DuplicateRecordError, DuplicateSampleError, CaseNotFoundError
from cg.store import models, Store

LOG = logging.getLogger(__name__)


class UploadObservationsAPI:

    """API to upload observations to LoqusDB."""

    def __init__(self, status_api: Store, hk_api: hk.HousekeeperAPI, loqus_api: loqus.LoqusdbAPI):
        self.status = status_api
        self.housekeeper = hk_api
        self.loqusdb = loqus_api

    def get_input(self, analysis_obj: models.Analysis) -> dict:
        """Fetch input data for an analysis to load observations."""

        input_data = {}

        analysis_date = analysis_obj.started_at or analysis_obj.completed_at
        hk_version = self.housekeeper.version(analysis_obj.family.internal_id, analysis_date)
        hk_vcf = self.housekeeper.files(version=hk_version.id, tags=["vcf-snv-research"]).first()
        hk_snv_gbcf = self.housekeeper.files(version=hk_version.id, tags=["snv-gbcf"]).first()
        hk_pedigree = self.housekeeper.files(version=hk_version.id, tags=["pedigree"]).first()

        input_data["sv_vcf"] = None
        if self.loqusdb.analysis_type == "wgs":
            hk_sv_vcf = self.housekeeper.files(
                version=hk_version.id, tags=["vcf-sv-research"]
            ).first()
            if hk_sv_vcf is None:
                raise FileNotFoundError("No file with vcf-sv-research tag in housekeeper")
            if not Path(hk_sv_vcf.full_path).exists():
                raise FileNotFoundError(f"{hk_sv_vcf.full_path} does not exist")
            input_data["sv_vcf"] = hk_sv_vcf.full_path

        hk_files = {"vcf-snv-research": hk_vcf, "snv-gbcf": hk_snv_gbcf, "pedigree": hk_pedigree}

        # Check if files exists. If hk returns None, or the path does not exist
        # FileNotFound error is raised
        for file_type, file in hk_files.items():
            if file is None:
                raise FileNotFoundError(f"No file with {file_type} tag in housekeeper")
            if not Path(file.full_path).exists():
                raise FileNotFoundError(f"{file.full_path} does not exist")

        input_data["family"] = analysis_obj.family.internal_id
        input_data["vcf"] = hk_vcf.full_path
        input_data["snv_gbcf"] = hk_snv_gbcf.full_path
        input_data["pedigree"] = hk_pedigree.full_path

        return input_data

    def upload(self, input_data: dict):
        """Upload data about genotypes for a family of samples."""

        try:
            existing_case = self.loqusdb.get_case(case_id=input_data["family"])

        # If CaseNotFoundError is raised, this should trigger the load method of loqusdb
        except CaseNotFoundError:

            duplicate = self.loqusdb.get_duplicate(input_data["snv_gbcf"])
            if duplicate:
                err_msg = f"Found duplicate {duplicate['ind_id']} in case {duplicate['case_id']}"
                raise DuplicateSampleError(err_msg)

            results = self.loqusdb.load(
                input_data["family"],
                input_data["pedigree"],
                input_data["vcf"],
                input_data["snv_gbcf"],
                vcf_sv_path=input_data["sv_vcf"],
            )
            LOG.info("parsed %s variants", results["variants"])

        else:
            log_msg = f"found existing family {existing_case['case_id']}, skipping observations"
            LOG.debug(log_msg)

    def process(self, analysis_obj: models.Analysis):
        """Process an upload observation counts for a case."""
        # check if some sample has already been uploaded
        for link in analysis_obj.family.links:
            if link.sample.loqusdb_id:
                raise DuplicateRecordError(f"{link.sample.internal_id} already in LoqusDB")
        results = self.get_input(analysis_obj)
        self.upload(results)
        case_obj = self.loqusdb.get_case(analysis_obj.family.internal_id)
        for link in analysis_obj.family.links:
            link.sample.loqusdb_id = str(case_obj["_id"])
        self.status.commit()

    @staticmethod
    def _all_samples(links: List[models.FamilySample]) -> bool:
        """Return True if all samples are external or sequenced inhouse."""
        return all((link.sample.sequenced_at or link.sample.is_external) for link in links)



"""Code to handle madeline input and output"""
import logging
from tempfile import NamedTemporaryFile, mkdtemp
from typing import List

from cg.utils import Process

LOG = logging.getLogger(__name__)


def make_ped(family_id: str, samples: List[dict]):
    """Yield lines that are used as madeline input."""
    columns = {
        "family": "FamilyId",
        "sample": "IndividualId",
        "sex": "Gender",
        "father": "Father",
        "mother": "Mother",
        "deceased": "Deceased",
        "proband": "Proband",
        "status": "Affected",
    }
    sex_gender = {"male": "M", "female": "F"}
    status_affected = {"affected": "Y", "unaffected": "N"}
    LOG.info("Generating madeline input lines")

    yield "\t".join(columns.values())

    for sample in samples:
        row = [
            family_id,
            sample["sample"],
            sex_gender.get(sample["sex"], "."),
            sample.get("father", "."),
            sample.get("mother", "."),
            "Y" if sample.get("deceased") else ".",
            "Y" if sample.get("proband") else ".",
            status_affected.get(sample.get("status"), "."),
        ]
        yield "\t".join(row)


def run(madeline_process: Process, ped_stream: List[str]):
    """Run madeline and generate a file with the results."""
    output_dir = mkdtemp()
    output_prefix = f"{output_dir}/madeline"
    out_path = f"{output_prefix}.xml"

    # write the input to a temp file
    with NamedTemporaryFile("w") as in_file:
        madeline_content = "\n".join(ped_stream)
        in_file.write(madeline_content)
        in_file.flush()
        madeline_call = [
            "--color",
            "--nolabeltruncation",
            "--outputprefix",
            output_prefix,
            in_file.name,
        ]
        madeline_process.run_command(parameters=madeline_call)

    with open(out_path, "r") as output:
        svg_content = output.read()

    # strip away the script tag
    script_tag = '<script type="text/javascript" xlink:href=' '"javascript/madeline.js"></script>'
    svg_content.replace(script_tag, "")

    with open(out_path, "w") as out_handle:
        out_handle.write(svg_content)

    return out_path



"""Fixtures for the meta tests"""

import pytest
from _pytest import tmpdir

from cg.apps.balsamic.fastq import FastqHandler as BalsamicFastqHandler
from cg.apps.crunchy import CrunchyAPI
from cg.apps.tb import TrailblazerAPI
from cg.apps.microsalt.fastq import FastqHandler as MicrosaltFastqHandler
from cg.meta.deliver import DeliverAPI
from cg.meta.workflow.mip import AnalysisAPI
from tests.mocks.hk_mock import MockFile


@pytest.yield_fixture(scope="function")
def analysis_store(base_store, analysis_family):
    """Setup a store instance for testing analysis API."""
    customer = base_store.customer("cust000")
    family = base_store.Family(
        name=analysis_family["name"],
        panels=analysis_family["panels"],
        internal_id=analysis_family["internal_id"],
        priority="standard",
    )
    family.customer = customer
    base_store.add(family)
    application_version = base_store.application("WGTPCFC030").versions[0]
    for sample_data in analysis_family["samples"]:
        sample = base_store.add_sample(
            name=sample_data["name"],
            sex=sample_data["sex"],
            internal_id=sample_data["internal_id"],
            ticket=sample_data["ticket_number"],
            reads=sample_data["reads"],
            capture_kit=sample_data["capture_kit"],
            data_analysis=sample_data["data_analysis"],
        )
        sample.family = family
        sample.application_version = application_version
        sample.customer = customer
        base_store.add(sample)
    base_store.commit()
    for sample_data in analysis_family["samples"]:
        sample_obj = base_store.sample(sample_data["internal_id"])
        link = base_store.relate_sample(
            family=family,
            sample=sample_obj,
            status=sample_data["status"],
            father=base_store.sample(sample_data["father"]) if sample_data.get("father") else None,
            mother=base_store.sample(sample_data["mother"]) if sample_data.get("mother") else None,
        )
        base_store.add(link)
    base_store.commit()
    yield base_store


class MockCrunchy(CrunchyAPI):

    pass


class MockLims:
    """Mock lims fixture"""

    lims = None

    def __init__(self):
        self.lims = self


class MockDeliver(DeliverAPI):
    def __init__(self):
        self.housekeeper = None
        self.lims = MockLims()

    def get_post_analysis_files(self, case: str, version, tags):

        if tags[0] == "mip-config":
            path = f"/mnt/hds/proj/bioinfo/bundles/{case}/2018-01-30/{case}_config.yaml"
        elif tags[0] == "sampleinfo":
            path = f"/mnt/hds/proj/bioinfo/bundles/{case}/2018-01-30/{case}_qc_sample_info.yaml"
        if tags[0] == "qcmetrics":
            path = f"/mnt/hds/proj/bioinfo/bundles/{case}/2018-01-30/{case}_qc_metrics.yaml"

        return [MockFile(path=path)]

    def get_post_analysis_case_files(self, case: str, version, tags):
        return ""

    def get_post_analysis_files_root_dir(self):
        return ""


class MockPath:
    def __init__(self, path):
        self.yaml = None

    def __call__(self, *args, **kwargs):
        return self

    def open(self):
        return dict()

    def joinpath(self, path):
        return ""


class MockLogger:
    warnings = []

    def warning(self, text, *interpolations):

        self.warnings.append(text % interpolations)

    def get_warnings(self) -> list:
        return self.warnings


class MockTB:
    _get_trending_raises_keyerror = False

    def __init__(self):
        """Needed to initialise mock variables"""
        self._make_config_was_called = False

    def get_trending(self, mip_config_raw: dict, qcmetrics_raw: dict, sampleinfo_raw: dict) -> dict:
        if self._get_trending_raises_keyerror:
            raise KeyError("mockmessage")

        # Returns: dict: parsed data
        ### Define output dict
        outdata = {
            "analysis_sex": {"ADM1": "female", "ADM2": "female", "ADM3": "female"},
            "family": "yellowhog",
            "duplicates": {"ADM1": 13.525, "ADM2": 12.525, "ADM3": 14.525},
            "genome_build": "hg19",
            "mapped_reads": {"ADM1": 98.8, "ADM2": 99.8, "ADM3": 97.8},
            "mip_version": "v4.0.20",
            "sample_ids": ["2018-20203", "2018-20204"],
            "genome_build": "37",
            "rank_model_version": "1.18",
        }

        return outdata

    def get_sampleinfo(self, analysis_obj):
        return ""

    def make_config(self, data, pipeline):
        """Mock the make_config"""
        self._make_config_was_called = True
        del pipeline
        return data


class MockBalsamicFastq(BalsamicFastqHandler):
    """Mock FastqHandler for analysis_api"""

    def __init__(self):
        super().__init__(config={"balsamic": {"root": tmpdir}})


class MockMicrosaltFastq(MicrosaltFastqHandler):
    """Mock FastqHandler for analysis_api"""

    def __init__(self):
        super().__init__(config={"microsalt": {"root": tmpdir}})


def safe_loader(path):
    """Mock function for loading yaml"""

    if path:
        pass

    return {
        "human_genome_build": {"version": ""},
        "recipe": {"rankvariant": {"rank_model": {"version": 1.18}}},
    }


@pytest.yield_fixture(scope="function")
def analysis_api(analysis_store, housekeeper_api, scout_api):
    """Setup an analysis API."""
    Path_mock = MockPath("")
    tb_mock = MockTB()
    deliver_mock = MockDeliver()
    deliver_mock.housekeeper = housekeeper_api

    _analysis_api = AnalysisAPI(
        db=analysis_store,
        hk_api=housekeeper_api,
        scout_api=scout_api,
        tb_api=tb_mock,
        lims_api=None,
        deliver_api=deliver_mock,
        yaml_loader=safe_loader,
        path_api=Path_mock,
        logger=MockLogger(),
        script="echo",
        pipeline="analyse rd_dna",
        conda_env="S_mip_rd-dna",
        root="/var/empty",
    )
    yield _analysis_api


@pytest.yield_fixture(scope="function")
def deliver_api(analysis_store, housekeeper_api, case_id, timestamp, helpers):
    """Fixture for deliver_api"""
    lims_mock = MockLims()
    hk_mock = housekeeper_api

    deliver_hk_bundle_data = {
        "name": case_id,
        "created": timestamp,
        "expires": timestamp,
        "files": [
            {"path": "/mock/path", "archive": False, "tags": ["case-tag"]},
            {"path": "/mock/path", "archive": False, "tags": ["sample-tag", "ADM1"]},
        ],
    }
    helpers.ensure_hk_bundle(hk_mock, deliver_hk_bundle_data)

    _api = DeliverAPI(
        db=analysis_store,
        hk_api=hk_mock,
        lims_api=lims_mock,
        case_tags=["case-tag"],
        sample_tags=["sample-tag"],
    )

    yield _api



""" fixtures for storing RNA analysis unit tests  """
import datetime
from pathlib import Path
import ruamel

import pytest


@pytest.fixture(scope="function")
def config_stream(files):
    """ config stream fixture """
    return {
        "rna_config_store": Path(files["rna_config_store"]).open(),
    }


@pytest.fixture(scope="function")
def config_raw(files_raw):
    """ config stream fixture """
    return files_raw["rna_config_store"]


@pytest.fixture(scope="function")
def config_data_rna():
    """ config data fixture for rna analyses"""
    return {
        "email": None,
        "case": "case_id",
        "samples": [{"id": "sample_id", "type": "wts"}],
        "is_dryrun": False,
        "out_dir": "tests/fixtures/apps/mip/rna/store",
        "priority": "low",
        "sampleinfo_path": "tests/fixtures/apps/mip/rna/store/case_qc_sample_info.yaml",
    }


@pytest.fixture(scope="function")
def config_data_dna():
    """ config data fixture for dna analyses"""
    return {
        "email": None,
        "case": "case_id",
        "samples": [{"id": "sample_id", "type": "wgs"}],
        "is_dryrun": False,
        "out_dir": "tests/fixtures/apps/mip/dna/store",
        "priority": "low",
        "sampleinfo_path": "tests/fixtures/apps/mip/dna/store/case_qc_sample_info.yaml",
    }


@pytest.fixture(scope="function")
def sampleinfo_data():
    """ sampleinfo data fixture """
    return {
        "date": datetime.datetime(2020, 3, 1),
        "is_finished": True,
        "case": "case_id",
        "version": "v8.2.2",
    }


@pytest.fixture(scope="function")
def rna_deliverables_raw(config_data_rna):
    """ raw_deliverables fixture """
    return ruamel.yaml.safe_load(
        Path(config_data_rna["out_dir"], "case_id_deliverables.yaml").open()
    )


@pytest.fixture(scope="function")
def dna_deliverables_raw(config_data_dna):
    """ raw_deliverables fixture """
    return ruamel.yaml.safe_load(
        Path(config_data_dna["out_dir"], "case_id_deliverables.yaml").open()
    )


@pytest.fixture(scope="function")
def bundle_data():
    """ bundle data fixture """
    return ruamel.yaml.safe_load(open("tests/fixtures/apps/mip/rna/store/bundle_data.yaml"))



"""Tests the Validator class"""
import pytest
from cg.meta.report.report_validator import ReportValidator

NON_REQUIRED_FIELDS_ON_ALL_SAMPLES = ["source", "data_analysis"]

NON_REQUIRED_FIELDS_ON_EXTERNAL_SAMPLES = [
    "million_read_pairs",
    "prepared_at",
    "prep_method",
    "sequencing_method",
    "sequenced_at",
]

REQUIRED_SEQUENCED_SAMPLE_FIELDS = [
    "received_at",
    "prep_method",
    "prepared_at",
    "sequencing_method",
    "sequenced_at",
    "million_read_pairs",
    "mapped_reads",
    "target_coverage",
    "target_completeness",
    "duplicates",
]

REQUIRED_ANALYSIS_SAMPLE_FIELDS = [
    "mapped_reads",
    "duplicates",
    "analysis_sex",
    "target_coverage",
    "target_completeness",
]

REQUIRED_REPORT_FIELDS = [
    "report_version",
    "case",
    "customer_name",
    "today",
    "panels",
    "customer_invoice_address",
    "scout_access",
    "accredited",
    "pipeline",
    "pipeline_version",
    "genome_build",
]

REQUIRED_GENERIC_SAMPLE_FIELDS = [
    "name",
    "internal_id",
    "sex",
    "status",
    "ticket",
    "application",
    "ordered_at",
    "received_at",
    "delivered_at",
    "processing_time",
    "analysis_sex",
]


@pytest.mark.parametrize("required_report_value_key", REQUIRED_REPORT_FIELDS)
def test_has_required_data_w_report_fields(report_api, required_report_value_key, report_store):
    # GIVEN complete delivery data missing only the data we check for value
    delivery_data = report_api._get_delivery_data(case_id="yellowhog")
    validator = ReportValidator(report_store)
    validator.has_required_data(delivery_data)
    assert not validator.get_missing_attributes()

    delivery_data[required_report_value_key] = None

    # WHEN checking that the data set has the required value
    has_required_data = validator.has_required_data(delivery_data)

    # THEN the indication on completeness should be false
    assert not has_required_data


@pytest.mark.parametrize("required_report_value_key", REQUIRED_REPORT_FIELDS)
def test_get_missing_data_w_report_fields(report_api, required_report_value_key, report_store):
    # GIVEN complete delivery data missing only the data we check for value
    delivery_data = report_api._get_delivery_data(case_id="yellowhog")
    validator = ReportValidator(report_store)
    assert validator.has_required_data(delivery_data)
    assert required_report_value_key not in validator.get_missing_attributes()

    delivery_data[required_report_value_key] = None
    validator.has_required_data(delivery_data)

    # WHEN getting missing attributes
    missing_attributes = validator.get_missing_attributes()

    # THEN the tested attribute should be one of the missing ones
    assert required_report_value_key in missing_attributes


@pytest.mark.parametrize("required_sample_value_key", REQUIRED_GENERIC_SAMPLE_FIELDS)
def test_has_required_data_w_normal_sample_fields(
    report_api, report_store, required_sample_value_key
):
    # GIVEN complete delivery data missing only the data we check for value on each sample
    delivery_data = report_api._get_delivery_data(case_id="yellowhog")
    validator = ReportValidator(report_store)
    assert validator.has_required_data(delivery_data)
    assert required_sample_value_key not in validator.get_missing_attributes()

    for sample in delivery_data["samples"]:
        sample[required_sample_value_key] = None

    # WHEN checking that the data set has the required value
    has_required_data = validator.has_required_data(delivery_data)

    # THEN the indication on completeness should be false
    assert not has_required_data


@pytest.mark.parametrize("required_sample_value_key", REQUIRED_SEQUENCED_SAMPLE_FIELDS)
def test_has_required_data_w_sequence_sample_fields(
    report_api, report_store, required_sample_value_key
):
    # GIVEN complete delivery data missing only the data we check for value on each non external
    # sample
    delivery_data = report_api._get_delivery_data(case_id="yellowhog")

    for sample in delivery_data["samples"]:
        assert not report_store.sample(
            sample["internal_id"]
        ).application_version.application.is_external

    validator = ReportValidator(report_store)
    assert validator.has_required_data(delivery_data)
    assert required_sample_value_key not in validator.get_missing_attributes()

    for sample in delivery_data["samples"]:
        sample[required_sample_value_key] = None

    # WHEN checking that the data set has the required value
    has_required_data = validator.has_required_data(delivery_data)

    # THEN the indication on completeness should be false
    assert not has_required_data


@pytest.mark.parametrize("required_sample_value_key", REQUIRED_ANALYSIS_SAMPLE_FIELDS)
def test_has_required_data_w_analysis_sample_fields(
    report_api, report_store, required_sample_value_key
):
    # GIVEN complete delivery data missing only the data we check for value on each non external
    # sample
    case_id = "yellowhog"
    delivery_data = report_api._get_delivery_data(case_id=case_id)
    for sample in delivery_data["samples"]:
        report_store.sample(sample["internal_id"]).data_analysis = "MIP"
    report_store.commit()

    validator = ReportValidator(report_store)
    assert validator.has_required_data(delivery_data)
    assert required_sample_value_key not in validator.get_missing_attributes()

    for sample in delivery_data["samples"]:
        sample[required_sample_value_key] = None

    # WHEN checking that the data set has the required value
    has_required_data = validator.has_required_data(delivery_data)

    # THEN the indication on completeness should be false
    assert not has_required_data


@pytest.mark.parametrize("required_sample_value_key", NON_REQUIRED_FIELDS_ON_ALL_SAMPLES)
def test_has_required_data_all_samples_non_required_fields(
    report_api, report_store, required_sample_value_key
):
    # GIVEN complete delivery data missing only the data we check for value on each sample
    delivery_data = report_api._get_delivery_data(case_id="yellowhog")

    validator = ReportValidator(report_store)
    assert validator.has_required_data(delivery_data)
    assert required_sample_value_key not in validator.get_missing_attributes()

    for sample in delivery_data["samples"]:
        sample[required_sample_value_key] = None

    # WHEN checking that the data set has the required value
    has_required_data = validator.has_required_data(delivery_data)

    # THEN the indication on completeness should be true
    assert required_sample_value_key not in validator.get_missing_attributes()
    assert has_required_data


@pytest.mark.parametrize("required_sample_value_key", NON_REQUIRED_FIELDS_ON_EXTERNAL_SAMPLES)
def test_has_required_data_external_samples_non_required_fields(
    report_api, report_store, required_sample_value_key
):
    # GIVEN complete delivery data missing only the data we check for value on each external sample
    case_id = "yellowhog"
    delivery_data = report_api._get_delivery_data(case_id=case_id)
    for sample in delivery_data["samples"]:
        report_store.sample(
            sample["internal_id"]
        ).application_version.application.is_external = True
    report_store.commit()

    for sample in delivery_data["samples"]:
        assert report_store.sample(
            sample["internal_id"]
        ).application_version.application.is_external

    validator = ReportValidator(report_store)
    assert validator.has_required_data(delivery_data)
    assert required_sample_value_key not in validator.get_missing_attributes()

    for sample in delivery_data["samples"]:
        sample[required_sample_value_key] = None

    # WHEN checking that the data set has the required value
    has_required_data = validator.has_required_data(delivery_data)

    # THEN the indication on completeness should be true
    assert required_sample_value_key not in validator.get_missing_attributes()
    assert has_required_data



"""Parse sample info files for RNA Cases """


def parse_sampleinfo_rna(data: dict) -> dict:
    """Parse MIP sample info file (RNA).

    Args:
        data (dict): raw YAML input from MIP qc sample info file (RNA)

    Returns:
        dict: parsed data
    """
    case = data["case"]
    qcmetrics_parse = data["recipe"]["qccollect_ar"]

    outdata = {
        "date": data["analysis_date"],
        "is_finished": data["analysisrunstatus"] == "finished",
        "case": case,
        "config_file_path": data["config_file_analysis"],
        "version": data["mip_version"],
        "pedigree_path": data["pedigree_minimal"],
        "bcftools_merge": data["recipe"]["bcftools_merge"]["path"],
        "multiqc_html": data["recipe"]["multiqc"][f"{case}_html"]["path"],
        "multiqc_json": data["recipe"]["multiqc"][f"{case}_json"]["path"],
        "qcmetrics_path": qcmetrics_parse if not isinstance(qcmetrics_parse, dict) else None,
        "vep_path": data["recipe"]["varianteffectpredictor"]["path"],
        "version_collect_ar_path": data["recipe"]["version_collect_ar"]["path"],
        "samples": [],
    }

    for sample_id, sample_data in data["sample"].items():

        sample = {
            "id": sample_id,
            "bam": sample_data["most_complete_bam"]["path"],
            "star_fusion": sample_data["recipe"]["star_fusion"]["path"],
            "bootstrap_vcf": get_multiple_paths(sample_data, "bootstrapann"),
            "gatk_asereadcounter": get_multiple_paths(sample_data, "gatk_asereadcounter"),
            "gatk_baserecalibration": get_multiple_paths(sample_data, "gatk_baserecalibration"),
            "gffcompare_ar": get_multiple_paths(sample_data, "gffcompare_ar"),
            "mark_duplicates": get_multiple_paths(sample_data, "markduplicates"),
            "salmon_quant": get_multiple_paths(sample_data, "salmon_quant"),
            "stringtie_ar": get_multiple_paths(sample_data, "stringtie_ar"),
        }

        outdata["samples"].append(sample)

    return outdata


def get_multiple_paths(sample_data: dict, path_key: str) -> list:
    """Get all paths to files of a given type. Use this method if the exact filename is not
    known beforehand.
    Args:
        sample_data (dict): YAML file containing qc sample info
        path_key (str)    : Type of file paths to retrieve

    Returns:
        list: paths to all files of given type
    """
    paths = [file_data["path"] for file_data in sample_data["recipe"][path_key].values()]

    return paths



"""Fixtures for cli compress functions"""

import pytest

from cg.apps.crunchy import CrunchyAPI
from cg.meta.compress import CompressAPI


class MockCompressAPI:
    """Mock out necessary functions for running the compress CLI functions"""

    def __init__(self):
        """initialize mock"""
        self.ntasks = 12
        self.mem = 50
        self.fastq_compression_success = True
        self.spring_decompression_success = True
        self.dry_run = False

    def set_dry_run(self, dry_run: bool):
        """Update dry run"""
        self.dry_run = dry_run

    def compress_fastq(self, sample_id: str, dry_run: bool = False):
        """Return if compression was succesfull"""
        _ = sample_id, dry_run
        return self.fastq_compression_success

    def decompress_spring(self, sample_id: str, dry_run: bool = False):
        """Return if decompression was succesfull"""
        _ = sample_id, dry_run
        return self.spring_decompression_success


@pytest.fixture(name="compress_api")
def fixture_compress_api():
    """Return a compress context"""
    return MockCompressAPI()


@pytest.yield_fixture(scope="function", name="real_crunchy_api")
def fixture_real_crunchy_api(crunchy_config_dict):
    """crunchy api fixture"""
    _api = CrunchyAPI(crunchy_config_dict)
    _api.set_dry_run(True)
    yield _api


@pytest.fixture(name="real_compress_api")
def fixture_real_compress_api(housekeeper_api, real_crunchy_api):
    """Return a compress context"""
    hk_api = housekeeper_api
    _api = CompressAPI(crunchy_api=real_crunchy_api, hk_api=hk_api)
    yield _api


@pytest.fixture(scope="function", name="real_populated_compress_fastq_api")
def fixture_real_populated_compress_fastq_api(real_compress_api, compress_hk_fastq_bundle, helpers):
    """Populated compress api fixture"""
    helpers.ensure_hk_bundle(real_compress_api.hk_api, compress_hk_fastq_bundle)

    return real_compress_api


@pytest.fixture(name="samples")
def fixture_samples():
    """Return a list of sample ids"""
    return ["sample1", "sample2", "sample3"]


# Store fixtures


class CaseInfo:
    """Holds information for creating a case"""

    def __init__(self, **kwargs):
        self.case_id = kwargs["case_id"]
        self.family_name = kwargs["family_name"]
        self.timestamp = kwargs["timestamp"]
        self.later_timestamp = kwargs["later_timestamp"]
        self.application_tag = kwargs["application_tag"]


@pytest.fixture(name="compress_case_info")
def fixture_compress_case_info(
    case_id,
    family_name,
    timestamp,
    later_timestamp,
    wgs_application_tag,
):
    """Returns a object with information about a case"""
    return CaseInfo(
        case_id=case_id,
        family_name=family_name,
        timestamp=timestamp,
        later_timestamp=later_timestamp,
        application_tag=wgs_application_tag,
    )


@pytest.fixture(name="populated_compress_store")
def fixture_populated_compress_store(store, helpers, compress_case_info, analysis_family):
    """Return a store populated with a completed analysis"""
    # Make sure that there is a family where anaylsis is completer
    helpers.ensure_family_from_dict(
        store,
        family_info=analysis_family,
        app_tag=compress_case_info.application_tag,
        ordered_at=compress_case_info.timestamp,
        completed_at=compress_case_info.later_timestamp,
    )

    return store


@pytest.fixture(name="populated_compress_multiple_store")
def fixture_populated_compress_multiple_store(
    store,
    helpers,
    compress_case_info,
    analysis_family,
):
    """Return a store populated with multiple completed analysis"""
    case_id = compress_case_info.case_id
    family_name = compress_case_info.family_name
    for number in range(10):
        analysis_family["internal_id"] = "_".join([str(number), case_id])
        analysis_family["name"] = "_".join([str(number), family_name])
        for ind, sample in enumerate(analysis_family["samples"]):
            analysis_family["samples"][ind]["internal_id"] = "_".join(
                [str(number), sample["internal_id"]]
            )
            if "father" in analysis_family["samples"][ind]:
                analysis_family["samples"][ind]["father"] = "_".join(
                    [str(number), analysis_family["samples"][ind]["father"]]
                )
            if "mother" in analysis_family["samples"][ind]:
                analysis_family["samples"][ind]["mother"] = "_".join(
                    [str(number), analysis_family["samples"][ind]["mother"]]
                )
        helpers.ensure_family_from_dict(
            store,
            family_info=analysis_family,
            app_tag=compress_case_info.application_tag,
            ordered_at=compress_case_info.timestamp,
            completed_at=compress_case_info.later_timestamp,
        )
    return store


# Context fixtures
@pytest.fixture(name="compress_context")
def fixture_base_compress_context(compress_api, store):
    """Return a compress context"""
    ctx = {"compress": compress_api, "db": store}
    return ctx


@pytest.fixture(name="store_fastq_context")
def fixture_store_fastq_context(compress_api, store):
    """Return a compress context"""
    ctx = {"compress": compress_api, "db": store}
    return ctx


@pytest.fixture(name="populated_multiple_compress_context")
def fixture_populated_multiple_compress_context(compress_api, populated_compress_multiple_store):
    """Return a compress context populated with a completed analysis"""
    # Make sure that there is a family where anaylis is completer
    return {"compress": compress_api, "db": populated_compress_multiple_store}


@pytest.fixture(name="populated_compress_context")
def fixture_populated_compress_context(compress_api, populated_compress_store):
    """Return a compress context populated with a completed analysis"""
    # Make sure that there is a family where analysis is completed
    return {"compress": compress_api, "db": populated_compress_store}


@pytest.fixture(name="real_populated_compress_context")
def fixture_real_populated_compress_context(
    real_populated_compress_fastq_api, populated_compress_store
):
    """Return a compress context populated with a completed analysis"""
    # Make sure that there is a family where analysis is completed
    return {"compress": real_populated_compress_fastq_api, "db": populated_compress_store}


# Bundle fixtures


@pytest.fixture(scope="function", name="sample")
def fixture_sample():
    """Return the sample id for first sample"""
    return "sample_1"


@pytest.fixture(scope="function", name="new_dir")
def fixture_new_dir(project_dir):
    """Return the path to a subdirectory"""
    new_dir = project_dir / "new_dir/"
    new_dir.mkdir()
    return new_dir


@pytest.fixture(name="spring_bundle")
def fixture_spring_bundle(project_dir, timestamp, sample):
    """Return a bundle with some spring info"""
    spring_file = project_dir / "file.spring"
    spring_file.touch()
    spring_meta_file = project_dir / "file.json"
    spring_meta_file.touch()
    hk_bundle_data = {
        "name": sample,
        "created": timestamp,
        "expires": timestamp,
        "files": [
            {"path": str(spring_file), "archive": False, "tags": [sample, "spring"]},
            {"path": str(spring_meta_file), "archive": False, "tags": [sample, "spring-metadata"]},
        ],
    }

    return hk_bundle_data


@pytest.fixture(name="spring_bundle_symlink_problem")
def fixture_spring_bundle_symlink_problem(project_dir, new_dir, timestamp, sample):
    """Setup an environment that is similar to the case we want to solve"""
    # Create spring files and faulty paths to spring files
    spring_file = project_dir / "file.spring"
    spring_file.touch()
    wrong_spring_file = new_dir / "file.spring"
    spring_meta_file = project_dir / "file.json"
    spring_meta_file.touch()
    wrong_spring_meta_file = new_dir / "file.json"
    hk_bundle_data = {
        "name": sample,
        "created": timestamp,
        "expires": timestamp,
        "files": [
            {"path": str(wrong_spring_file), "archive": False, "tags": [sample, "spring"]},
            {
                "path": str(wrong_spring_meta_file),
                "archive": False,
                "tags": [sample, "spring-metadata"],
            },
        ],
    }

    return hk_bundle_data


@pytest.fixture(name="symlinked_fastqs")
def fixture_symlinked_fastqs(project_dir, new_dir):
    """Setup an environment that is similar to the case we want to solve"""
    fastq_first = project_dir / "first.fastq.gz"
    fastq_second = project_dir / "second.fastq.gz"

    symlinked_first = new_dir / "first.fastq.gz"
    symlinked_first.symlink_to(fastq_first)
    symlinked_second = new_dir / "second.fastq.gz"
    symlinked_second.symlink_to(fastq_second)
    return {
        "fastq_first": fastq_first,
        "fastq_second": fastq_second,
        "symlinked_first": symlinked_first,
        "symlinked_second": symlinked_second,
    }



"""This script tests the cli methods to set families to status-db"""

from cg.cli.set.family import family
from cg.store import Store

SUCCESS = 0


def test_set_family_without_options(cli_runner, base_context, base_store: Store, helpers):
    """Test to set a family using only the required arguments"""
    # GIVEN a database with a family
    family_id = helpers.add_family(base_store).internal_id
    assert base_store.Family.query.count() == 1

    # WHEN setting a family
    result = cli_runner.invoke(family, [family_id], obj=base_context)

    # THEN then it should abort
    assert result.exit_code != SUCCESS


def test_set_family_bad_family(cli_runner, base_context):
    """Test to set a family using a non-existing family """
    # GIVEN an empty database

    # WHEN setting a family
    family_id = "dummy_name"
    result = cli_runner.invoke(family, [family_id], obj=base_context)

    # THEN then it should complain on missing family
    assert result.exit_code != SUCCESS


def test_set_family_bad_panel(cli_runner, base_context, base_store: Store, helpers):
    """Test to set a family using a non-existing panel"""
    # GIVEN a database with a family

    # WHEN setting a family
    panel_id = "dummy_panel"
    family_id = helpers.add_family(base_store).internal_id
    result = cli_runner.invoke(family, [family_id, "--panel", panel_id], obj=base_context)

    # THEN then it should complain in missing panel instead of setting a value
    assert result.exit_code != SUCCESS
    assert panel_id not in base_store.Family.query.first().panels


def test_set_family_panel(cli_runner, base_context, base_store: Store, helpers):
    """Test to set a family using an existing panel"""
    # GIVEN a database with a family and a panel not yet added to the family
    panel_id = helpers.ensure_panel(base_store, "a_panel").name
    family_id = helpers.add_family(base_store).internal_id
    assert panel_id not in base_store.Family.query.first().panels

    # WHEN setting a panel of a family
    result = cli_runner.invoke(family, [family_id, "--panel", panel_id], obj=base_context)

    # THEN then it should set panel on the family
    assert result.exit_code == SUCCESS
    assert panel_id in base_store.Family.query.first().panels


def test_set_family_priority(cli_runner, base_context, base_store: Store, helpers):
    """Test that the added family gets the priority we send in"""
    # GIVEN a database with a family
    family_id = helpers.add_family(base_store).internal_id
    priority = "priority"
    assert base_store.Family.query.first().priority_human != priority

    # WHEN setting a family

    result = cli_runner.invoke(family, [family_id, "--priority", priority], obj=base_context)

    # THEN then it should have been set
    assert result.exit_code == SUCCESS
    assert base_store.Family.query.count() == 1
    assert base_store.Family.query.first().priority_human == priority


def test_set_family_customer(cli_runner, base_context, base_store: Store, helpers):
    """Test to set a family using an existing customer"""
    # GIVEN a database with a family and a customer not yet on the family
    customer_id = helpers.ensure_customer(base_store, customer_id="a_customer").internal_id
    case = helpers.add_family(base_store)
    assert customer_id != case.customer.internal_id

    # WHEN setting a customer of a family
    result = cli_runner.invoke(
        family, [case.internal_id, "--customer-id", customer_id], obj=base_context
    )

    # THEN then it should set customer on the family
    assert result.exit_code == SUCCESS
    assert customer_id == case.customer.internal_id



"""Test methods for compressing FASTQ"""
import json
import logging

from cg.apps.crunchy import CrunchyAPI


def test_get_spring_metadata(spring_metadata_file, crunchy_config_dict):
    """Test the method that fetches the SPRING metadata from a file"""
    # GIVEN a SPRING path and the path to a populated SPRING metadata file
    assert spring_metadata_file.is_file()
    assert spring_metadata_file.exists()
    # GIVEN a crunchy API
    crunchy_api = CrunchyAPI(crunchy_config_dict)

    # WHEN fetching the content of the file
    parsed_content = crunchy_api.get_spring_metadata(spring_metadata_file)

    # THEN assert information about the three files is there
    assert len(parsed_content) == 3
    # THEN assert a list with the files is returned
    assert isinstance(parsed_content, list)


def test_get_spring_archive_files(spring_metadata):
    """Test the method that sorts the SPRING metadata into a dictionary"""
    # GIVEN a SPRING metadata content in its raw format
    assert isinstance(spring_metadata, list)

    # WHEN fetching the content of the file
    parsed_content = CrunchyAPI.get_spring_archive_files(spring_metadata)

    # THEN assert information about the three files is there
    assert len(parsed_content) == 3
    # THEN assert a dictionary with the files is returned
    assert isinstance(parsed_content, dict)


def test_get_spring_metadata_malformed_info(
    spring_metadata_file, spring_metadata, crunchy_config_dict
):
    """Test the method that fetches the SPRING metadata from a file when file is malformed"""
    # GIVEN a SPRING metadata file with missing information
    spring_metadata[0].pop("path")
    with open(spring_metadata_file, "w") as outfile:
        outfile.write(json.dumps(spring_metadata))

    # GIVEN a crunchy API
    crunchy_api = CrunchyAPI(crunchy_config_dict)

    # WHEN fetching the content of the file
    parsed_content = crunchy_api.get_spring_metadata(spring_metadata_file)

    # THEN assert that None is returned to indicate failure
    assert parsed_content is None


def test_get_spring_metadata_wrong_number_files(
    spring_metadata_file, spring_metadata, crunchy_config_dict
):
    """Test the method that fetches the SPRING metadata from a file when a file is missing"""
    # GIVEN a SPRING metadata file with missing file
    spring_metadata = spring_metadata[1:]
    with open(spring_metadata_file, "w") as outfile:
        outfile.write(json.dumps(spring_metadata))

    # GIVEN a crunchy API
    crunchy_api = CrunchyAPI(crunchy_config_dict)

    # WHEN fetching the content of the file
    parsed_content = crunchy_api.get_spring_metadata(spring_metadata_file)

    # THEN assert that None is returned to indicate failure
    assert parsed_content is None


def test_fastq_to_spring_sbatch(crunchy_config_dict, compression_object, sbatch_process, caplog):
    """Test fastq_to_spring method"""
    caplog.set_level(logging.DEBUG)
    # GIVEN a crunchy-api, and FASTQ paths

    crunchy_api = CrunchyAPI(crunchy_config_dict)
    crunchy_api.process = sbatch_process
    spring_path = compression_object.spring_path
    log_path = crunchy_api.get_log_dir(spring_path)
    run_name = compression_object.run_name
    sbatch_path = crunchy_api.get_sbatch_path(log_path, "fastq", run_name)

    # GIVEN that the sbatch file does not exist
    assert not sbatch_path.is_file()

    # WHEN calling fastq_to_spring on FASTQ files
    crunchy_api.fastq_to_spring(compression_obj=compression_object)

    # THEN assert that the sbatch file was created
    assert sbatch_path.is_file()


def test_spring_to_fastq(compression_object, spring_metadata_file, crunchy_config_dict, mocker):
    """Test SPRING to FASTQ method

    Test to decompress SPRING to FASTQ. This test will make sure that the correct sbatch content
    was submitted to the Process api
    """
    # GIVEN a crunchy-api given an existing SPRING metadata file
    assert spring_metadata_file.exists()
    mocker_submit_sbatch = mocker.patch.object(CrunchyAPI, "_submit_sbatch")
    crunchy_api = CrunchyAPI(crunchy_config_dict)

    # WHEN calling bam_to_cram method on bam-path
    crunchy_api.spring_to_fastq(compression_obj=compression_object)

    # THEN _submit_sbatch method is called
    mocker_submit_sbatch.assert_called()



"""Code for validating an upload via CLI"""

import click

from cg.apps import coverage as coverage_app

from .utils import suggest_cases_to_upload


@click.command()
@click.argument("family_id", required=False)
@click.pass_context
def validate(context, family_id):
    """Validate a family of samples."""

    click.echo(click.style("----------------- VALIDATE --------------------"))

    if not family_id:
        suggest_cases_to_upload(context)
        context.abort()

    family_obj = context.obj["status"].family(family_id)
    chanjo_api = coverage_app.ChanjoAPI(context.obj)
    chanjo_samples = []
    for link_obj in family_obj.links:
        sample_id = link_obj.sample.internal_id
        chanjo_sample = chanjo_api.sample(sample_id)
        if chanjo_sample is None:
            click.echo(click.style(f"upload coverage for {sample_id}", fg="yellow"))
            continue
        chanjo_samples.append(chanjo_sample)

    if not chanjo_samples:
        return

    coverage_results = chanjo_api.omim_coverage(chanjo_samples)
    for link_obj in family_obj.links:
        sample_id = link_obj.sample.internal_id
        if sample_id in coverage_results:
            completeness = coverage_results[sample_id]["mean_completeness"]
            mean_coverage = coverage_results[sample_id]["mean_coverage"]
            click.echo(f"{sample_id}: {mean_coverage:.2f}X - {completeness:.2f}%")
        else:
            click.echo(f"{sample_id}: sample not found in chanjo", color="yellow")



OPTIONAL_KEYS = (
    "container_name",
    "quantity",
    "volume",
    "concentration",
    "status",
    "comment",
    "capture_kit",
    "mother",
    "father",
)


def parse_json(indata: dict) -> dict:
    """Parse JSON from LIMS export."""
    data = {
        "project_type": "mip",
        "customer": indata["customer"].lower(),
        "name": indata.get("name"),
        "comment": indata.get("comment"),
        "items": [],
    }
    families = {}
    for sample_data in indata["samples"]:
        if sample_data["family_name"] not in families:
            families[sample_data["family_name"]] = []
        families[sample_data["family_name"]].append(sample_data)

    for family_name, samples in families.items():
        family_data = {"name": family_name, "samples": []}
        require_qcoks = set(sample.get("require_qcok", False) for sample in samples)
        if True in require_qcoks:
            family_data["require_qcok"] = True

        priorities = set(sample["priority"].lower() for sample in samples)
        if "express" in priorities:
            family_data["priority"] = "express"
        elif "priority" in priorities:
            family_data["priority"] = "priority"
        else:
            family_data["priority"] = priorities.pop()

        panels = set()
        for sample in samples:
            panels.update(sample["panels"])
            sample_data = {
                "name": sample["name"],
                "sex": sample["sex"],
                "application": sample["application"],
                "source": sample["source"],
                "container": sample["container"],
                "data_analysis": sample.get("data_analysis", "MIP"),
            }
            well_position_raw = sample.get("well_position")
            if well_position_raw:
                sample_data["well_position"] = (
                    ":".join(well_position_raw)
                    if ":" not in well_position_raw
                    else well_position_raw
                )
            for key in OPTIONAL_KEYS:
                if sample.get(key):
                    sample_data[key] = sample[key]
            family_data["samples"].append(sample_data)
        family_data["panels"] = list(panels)
        data["items"].append(family_data)

    return data



"""API to the deployment tool Shipping"""

import logging
from pathlib import Path
from typing import Dict

from cg.utils import Process

LOG = logging.getLogger(__name__)


class ShippingAPI:
    """Class to communicate with the tool shipping

    Shipping is used as a unified tool for deploying in the Clinical Genomics environments
    """

    def __init__(self, config: Dict[str, str]):
        self.config = config
        self.host_config = config["host_config"]
        self.binary_path = config["binary_path"]
        self.process = Process(
            binary=str(self.binary_path), config=self.host_config, config_parameter="--host-config"
        )
        self.dry_run = False

    def set_dry_run(self, dry_run: bool) -> None:
        """Update dry run"""
        LOG.info("Set dry run to %s", dry_run)
        self.dry_run = dry_run

    def deploy(self, app_name: str, app_config: Path = None):
        """Command to deploy a tool according to the specifications in the config files"""
        LOG.info("Deploying the %s software", app_name)
        deploy_args = []
        if app_config:
            deploy_args.extend(["--app-config", str(app_config)])
        else:
            deploy_args.extend(["--tool-name", app_name])

        deploy_args.append("deploy")
        self.process.run_command(deploy_args, dry_run=self.dry_run)
        for line in self.process.stderr_lines():
            LOG.info(line)



"""Code for uploading genotype data via CLI"""
import click
import logging

from cg.meta.upload.gt import UploadGenotypesAPI

from .utils import suggest_cases_to_upload

LOG = logging.getLogger(__name__)


@click.command()
@click.option("-r", "--re-upload", is_flag=True, help="re-upload existing analysis")
@click.argument("family_id", required=False)
@click.pass_context
def genotypes(context, re_upload, family_id):
    """Upload genotypes from an analysis to Genotype."""

    click.echo(click.style("----------------- GENOTYPES -------------------"))

    if not family_id:
        suggest_cases_to_upload(context)
        context.abort()

    gt_api = context.obj["genotype_api"]
    hk_api = context.obj["housekeeper_api"]
    status_api = context.obj["status"]
    family_obj = status_api.family(family_id)
    upload_genotypes_api = UploadGenotypesAPI(hk_api=hk_api, gt_api=gt_api)
    results = upload_genotypes_api.data(family_obj.analyses[0])

    if results:
        upload_genotypes_api.upload(results, replace=re_upload)



"""This script tests the cli methods to add families to status-db"""

from cg.store import Store


def test_add_relationship_required(invoke_cli, disk_store: Store, helpers):
    """Test to add a relationship using only the required arguments"""
    # GIVEN a database with a sample and an family
    sample = helpers.add_sample(disk_store)
    sample_id = sample.internal_id
    family = helpers.add_family(disk_store)
    family_id = family.internal_id
    status = "affected"

    # WHEN adding a relationship
    db_uri = disk_store.uri

    result = invoke_cli(
        [
            "--database",
            db_uri,
            "add",
            "relationship",
            family_id,
            sample_id,
            "-s",
            status,
        ]
    )

    # THEN then it should be added
    assert result.exit_code == 0
    assert disk_store.FamilySample.query.count() == 1
    assert disk_store.FamilySample.query.first().family.internal_id == family_id
    assert disk_store.FamilySample.query.first().sample.internal_id == sample_id


def test_add_relationship_bad_sample(invoke_cli, disk_store: Store, helpers):
    """Test to add a relationship using a non-existing sample"""
    # GIVEN an empty database

    # WHEN adding a relationship
    db_uri = disk_store.uri
    family = helpers.add_family(disk_store)
    family_id = family.internal_id
    sample_id = "dummy_sample"
    status = "affected"
    result = invoke_cli(
        [
            "--database",
            db_uri,
            "add",
            "relationship",
            family_id,
            sample_id,
            "-s",
            status,
        ]
    )

    # THEN then it should complain on missing sample instead of adding a relationship
    assert result.exit_code == 1
    assert disk_store.FamilySample.query.count() == 0


def test_add_relationship_bad_family(invoke_cli, disk_store: Store, helpers):
    """Test to add a relationship using a non-existing family"""
    # GIVEN a database with a sample

    # WHEN adding a relationship
    db_uri = disk_store.uri
    family_id = "dummy_family"
    sample = helpers.add_sample(disk_store)
    sample_id = sample.internal_id

    status = "affected"
    result = invoke_cli(
        [
            "--database",
            db_uri,
            "add",
            "relationship",
            family_id,
            sample_id,
            "-s",
            status,
        ]
    )

    # THEN then it should complain in missing family instead of adding a relationship
    assert result.exit_code == 1
    assert disk_store.FamilySample.query.count() == 0


def test_add_relationship_bad_status(invoke_cli, disk_store: Store, helpers):
    """Test that the added relationship get the status we send in"""
    # GIVEN a database with a sample and an family

    # WHEN adding a relationship
    db_uri = disk_store.uri

    sample = helpers.add_sample(disk_store)
    sample_id = sample.internal_id
    family = helpers.add_family(disk_store)
    family_id = family.internal_id
    status = "dummy_status"

    result = invoke_cli(
        [
            "--database",
            db_uri,
            "add",
            "relationship",
            family_id,
            sample_id,
            "-s",
            status,
        ]
    )

    # THEN then it should complain on bad status instead of adding a relationship
    assert result.exit_code == 2
    assert disk_store.FamilySample.query.count() == 0


def test_add_relationship_mother(invoke_cli, disk_store: Store, helpers):
    """Test to add a relationship with a mother"""
    # GIVEN a database with a sample and an family
    sample = helpers.add_sample(disk_store)
    sample_id = sample.internal_id
    mother = helpers.add_sample(disk_store, sample_id="mother")
    mother_id = mother.internal_id
    family = helpers.add_family(disk_store)
    family_id = family.internal_id
    status = "affected"

    # WHEN adding a relationship
    db_uri = disk_store.uri

    result = invoke_cli(
        [
            "--database",
            db_uri,
            "add",
            "relationship",
            family_id,
            sample_id,
            "-s",
            status,
            "-m",
            mother_id,
        ]
    )

    # THEN then it should be added
    assert result.exit_code == 0
    assert disk_store.FamilySample.query.count() == 1
    assert disk_store.FamilySample.query.first().mother.internal_id == mother_id


def test_add_relationship_bad_mother(invoke_cli, disk_store: Store, helpers):
    """Test to add a relationship using a non-existing mother"""
    # GIVEN a database with a sample and an family
    sample = helpers.add_sample(disk_store)
    sample_id = sample.internal_id
    mother_id = "dummy_mother"
    family = helpers.add_family(disk_store)
    family_id = family.internal_id
    status = "affected"

    # WHEN adding a relationship
    db_uri = disk_store.uri

    result = invoke_cli(
        [
            "--database",
            db_uri,
            "add",
            "relationship",
            family_id,
            sample_id,
            "-s",
            status,
            "-m",
            mother_id,
        ]
    )

    # THEN then it should not be added
    assert result.exit_code == 1
    assert disk_store.FamilySample.query.count() == 0


def test_add_relationship_father(invoke_cli, disk_store: Store, helpers):
    """Test to add a relationship using a father"""
    # GIVEN a database with a sample and an family
    sample = helpers.add_sample(disk_store)
    sample_id = sample.internal_id

    father = helpers.add_sample(disk_store, sample_id="father", gender="male")
    father_id = father.internal_id

    family = helpers.add_family(disk_store)
    family_id = family.internal_id

    status = "affected"

    # WHEN adding a relationship
    db_uri = disk_store.uri

    result = invoke_cli(
        [
            "--database",
            db_uri,
            "add",
            "relationship",
            family_id,
            sample_id,
            "-s",
            status,
            "-f",
            father_id,
        ]
    )

    # THEN then it should be added
    assert result.exit_code == 0
    assert disk_store.FamilySample.query.count() == 1
    assert disk_store.FamilySample.query.first().father.internal_id == father_id


def test_add_relationship_bad_father(invoke_cli, disk_store: Store, helpers):
    """Test to add a relationship using a non-existing father"""
    # GIVEN a database with a sample and an family
    sample = helpers.add_sample(disk_store)
    sample_id = sample.internal_id

    father_id = "bad_father"
    family = helpers.add_family(disk_store)
    family_id = family.internal_id

    status = "affected"

    # WHEN adding a relationship
    db_uri = disk_store.uri

    result = invoke_cli(
        [
            "--database",
            db_uri,
            "add",
            "relationship",
            family_id,
            sample_id,
            "-s",
            status,
            "-f",
            father_id,
        ]
    )

    # THEN then it should not be added
    assert result.exit_code == 1
    assert disk_store.FamilySample.query.count() == 0



from cg.meta.orders.lims import LimsHandler


def test_to_lims_mip(mip_order_to_submit):
    # GIVEN a scout order for a trio
    # WHEN parsing the order to format for LIMS import
    samples = LimsHandler.to_lims(customer="cust003", samples=mip_order_to_submit["samples"])
    # THEN it should list all samples
    assert len(samples) == 4
    # ... and determine the container, container name, and well position
    assert set(sample["container"] for sample in samples) == set(["96 well plate"])
    container_names = set(
        sample["container_name"] for sample in samples if sample["container_name"]
    )
    assert container_names == set(["CMMS"])
    assert samples[0]["well_position"] == "A:1"
    # ... and pick out relevant UDFs
    first_sample = samples[0]
    assert first_sample["udfs"]["family_name"] == "family1"
    assert first_sample["udfs"]["priority"] == "standard"
    assert first_sample["udfs"]["application"] == "WGTPCFC030"
    assert first_sample["udfs"]["source"] == "tissue (fresh frozen)"
    assert first_sample["udfs"]["quantity"] == "220"
    assert first_sample["udfs"]["require_qcok"] is True
    assert first_sample["udfs"]["customer"] == "cust003"
    assert isinstance(samples[1]["udfs"]["comment"], str)


def test_to_lims_external(external_order_to_submit):
    # GIVEN an external order for two samples
    # WHEN parsing the order to format for LIMS
    samples = LimsHandler.to_lims(customer="dummyCust", samples=external_order_to_submit["samples"])
    # THEN should "work"
    assert len(samples) == 2
    # ... and make up a container for each sample
    assert samples[0]["container"] == "Tube"


def test_to_lims_fastq(fastq_order_to_submit):
    # GIVEN a fastq order for two samples; normal vs. tumour
    # WHEN parsing the order to format for LIMS
    samples = LimsHandler.to_lims(customer="dummyCust", samples=fastq_order_to_submit["samples"])
    # THEN should "work"
    assert len(samples) == 2
    # ... and pick out relevant UDF values
    assert samples[0]["udfs"]["tumour"] is False
    assert samples[1]["udfs"]["tumour"] is True


def test_to_lims_rml(rml_order_to_submit):
    # GIVEN a rml order for three samples
    # WHEN parsing for LIMS
    samples = LimsHandler.to_lims(customer="dummyCust", samples=rml_order_to_submit["samples"])
    # THEN it should "work"
    assert len(samples) == 2
    # ... and pick out relevant UDFs
    first_sample = samples[0]
    assert first_sample["udfs"]["pool"] == "pool-1"
    assert first_sample["udfs"]["volume"] == "30"
    assert first_sample["udfs"]["concentration"] == "5"
    assert first_sample["udfs"]["index"] == "TruSeq DNA HT Dual-index (D7-D5)"
    assert first_sample["udfs"]["index_number"] == "1"


def test_to_lims_microbial(microbial_order_to_submit):
    # GIVEN a microbial order for three samples
    # WHEN parsing for LIMS
    samples = LimsHandler.to_lims(customer="cust000", samples=microbial_order_to_submit["samples"])
    # THEN it should "work"
    assert len(samples) == 5
    # ... and pick out relevant UDFs
    first_sample = samples[0]
    assert first_sample["udfs"]["priority"] == "research"
    assert first_sample["udfs"]["organism"] == "M.upium"
    assert first_sample["udfs"]["reference_genome"] == "NC_111"
    assert (
        first_sample["udfs"]["extraction_method"] == "MagNaPure 96 (contact Clinical Genomics "
        "before submission)"
    )


def test_to_lims_balsamic(balsamic_order_to_submit):

    # GIVEN a cancer order for a sample
    # WHEN parsing the order to format for LIMS import
    samples = LimsHandler.to_lims(customer="cust000", samples=balsamic_order_to_submit["samples"])
    # THEN it should list all samples

    assert len(samples) == 1
    # ... and determine the container, container name, and well position

    container_names = set(
        sample["container_name"] for sample in samples if sample["container_name"]
    )

    # ... and pick out relevant UDFs
    first_sample = samples[0]
    assert first_sample["name"] == "s1"
    assert set(sample["container"] for sample in samples) == set(["96 well plate"])
    assert first_sample["udfs"]["data_analysis"] == "Balsamic"
    assert first_sample["udfs"]["application"] == "WGTPCFC030"
    assert first_sample["udfs"]["sex"] == "M"
    assert first_sample["udfs"]["family_name"] == "family1"
    assert first_sample["udfs"]["customer"] == "cust000"
    assert first_sample["udfs"]["require_qcok"] is False
    assert first_sample["udfs"]["source"] == "blood"
    assert first_sample["udfs"]["priority"] == "standard"

    assert container_names == set(["p1"])
    assert samples[0]["well_position"] == "A:1"

    assert first_sample["udfs"]["tumour"] is True
    assert first_sample["udfs"]["capture_kit"] == "LymphoMATIC"
    assert first_sample["udfs"]["tumour_purity"] == "75"

    assert first_sample["udfs"]["formalin_fixation_time"] == "1"
    assert first_sample["udfs"]["post_formalin_fixation_time"] == "2"
    assert first_sample["udfs"]["tissue_block_size"] == "small"

    assert first_sample["udfs"]["quantity"] == "2"
    assert first_sample["udfs"]["comment"] == "comment"


def test_to_lims_mip_balsamic(mip_balsamic_order_to_submit):
    # GIVEN a scout order for a trio

    # WHEN parsing the order to format for LIMS import
    samples = LimsHandler.to_lims(
        customer="cust003", samples=mip_balsamic_order_to_submit["samples"]
    )

    # THEN it should list all samples
    assert len(samples) == 4
    # ... and determine the container, container name, and well position
    assert set(sample["container"] for sample in samples) == set(["96 well plate"])
    container_names = set(
        sample["container_name"] for sample in samples if sample["container_name"]
    )
    assert container_names == set(["CMMS"])
    assert samples[0]["well_position"] == "A:1"
    # ... and pick out relevant UDFs
    first_sample = samples[0]
    assert first_sample["udfs"]["family_name"] == "family1"
    assert first_sample["udfs"]["priority"] == "standard"
    assert first_sample["udfs"]["application"] == "WGTPCFC030"
    assert first_sample["udfs"]["source"] == "tissue (fresh frozen)"
    assert first_sample["udfs"]["quantity"] == "220"
    assert first_sample["udfs"]["require_qcok"] is True
    assert first_sample["udfs"]["customer"] == "cust003"
    assert isinstance(samples[1]["udfs"]["comment"], str)

    assert first_sample["udfs"]["tumour"] is True
    assert first_sample["udfs"]["capture_kit"] == "LymphoMATIC"
    assert first_sample["udfs"]["tumour_purity"] == "75"

    assert first_sample["udfs"]["formalin_fixation_time"] == "1"
    assert first_sample["udfs"]["post_formalin_fixation_time"] == "2"
    assert first_sample["udfs"]["tissue_block_size"] == "large"



"""Tests for delivery API"""


def test_get_post_analysis_files(analysis_family, deliver_api):

    # GIVEN a case which exist as bundle in hk
    # GIVEN corresponding houskeeper tags for those files
    case = analysis_family["internal_id"]
    deliver_tags = ["case-tag"]

    # WHEN we call get_post_analysis_files with matching case and file-tags and version is no version
    version = None
    bundle_latest_files = deliver_api.get_post_analysis_files(case, version, deliver_tags)

    ## THEN housekeeper files should be returned
    assert bundle_latest_files


def test_get_post_analysis_files_version(analysis_family, deliver_api):

    # GIVEN a case which exist as bundle in hk
    # GIVEN corresponding houskeeper tags for those files
    case = analysis_family["internal_id"]
    deliver_tags = ["case-tag"]

    # WHEN we call get_post_analysis_files with matching case and file-tags and version is supplied
    version = "a_date"

    bundle_version_files = deliver_api.get_post_analysis_files(case, version, deliver_tags)

    # THEN housekeeper files for those tags with the version should be returned
    assert bundle_version_files


def test_get_post_analysis_case_files(analysis_family, deliver_api):

    # GIVEN a case which exist as bundle in hk
    # GIVEN corresponding case houskeeper tags for those files
    case = analysis_family["internal_id"]
    deliver_tags = ["case-tag"]

    # WHEN we call get_post_analysis_case_files with matching case and case file-tags and version
    version = None
    case_files = deliver_api.get_post_analysis_case_files(case, version, deliver_tags)

    # THEN we should get those matching files back
    assert case_files


def test_get_post_analysis_case_files_only_sample_tags(analysis_family, deliver_api):

    # GIVEN a case which exist as bundle in hk
    # GIVEN corresponding sample id houskeeper tags for those files
    case = analysis_family["internal_id"]
    deliver_tags = ["sample-tag"]
    hk_api = deliver_api.hk_api
    hk_api.add_missing_tag(deliver_tags[0])

    # WHEN we call get_post_analysis_case_files with matching case and sample file-tags and version
    version = None
    case_files = deliver_api.get_post_analysis_case_files(case, version, deliver_tags)

    # THEN we should get those matching files back
    assert not case_files


def test_get_post_analysis_sample_files(analysis_family, deliver_api):

    # GIVEN a case which exist as bundle in hk
    # GIVEN corresponding houskeeper tags for those files
    case = analysis_family["internal_id"]
    sample = analysis_family["samples"][0]["internal_id"]
    deliver_tags = ["sample-tag"]

    # WHEN we call get_post_analysis_sample_files with matching case and file-tags and version
    version = None
    sample_files = deliver_api.get_post_analysis_sample_files(case, sample, version, deliver_tags)

    # THEN we should get those matching files back
    assert sample_files


def test_get_post_analysis_sample_files_only_case_tags(analysis_family, deliver_api):

    # GIVEN a case which exist as bundle in hk
    # GIVEN corresponding houskeeper tags for those files
    case = analysis_family["internal_id"]
    sample = analysis_family["samples"][0]["internal_id"]
    deliver_tags = ["case-tag"]
    hk_api = deliver_api.hk_api
    hk_api.add_missing_tag(deliver_tags[0])

    # WHEN we call get_post_analysis_sample_files with matching case and file-tags and version
    version = None
    sample_files = deliver_api.get_post_analysis_sample_files(case, sample, version, deliver_tags)

    # THEN we should get those matching files back
    assert not sample_files


def test_get_post_analysis_files_root_dir(analysis_family, deliver_api):

    ## WHEN we call get_post_analysis_files_root_dir
    root_dir = deliver_api.get_post_analysis_files_root_dir()

    ## THEN we should get a path back
    assert root_dir



""" Microsalt specific functionality for storing files in Housekeeper """
import datetime as dt
import logging
import os
import ruamel.yaml

from _io import TextIOWrapper

from cg.constants import HK_TAGS, MICROSALT_TAGS
from cg.meta.store import base as store_base

LOG = logging.getLogger(__name__)


def add_microbial_analysis(config_stream: TextIOWrapper) -> dict:
    """Gather information from microSALT analysis to store."""

    deliverables = ruamel.yaml.safe_load(config_stream)
    analysis_date = _get_date_from_deliverables_path(config_stream.name)
    new_bundle = build_microbial_bundle(deliverables, analysis_date)

    return new_bundle


def build_microbial_bundle(deliverables: dict, analysis_date: dt.date) -> dict:
    """Create a new bundle to store in Housekeeper"""

    project_name = _get_microbial_name(deliverables)
    files = store_base.deliverables_files(deliverables, analysis_type="microbial")

    data = {
        "name": str(project_name),
        "created": analysis_date,
        "files": files,
    }
    return data


def microbial_deliverables_files(deliverables: dict) -> list:
    """Get all deliverable files from the pipeline"""

    pipeline_tags = HK_TAGS["microbial"]
    analysis_type_tags = MICROSALT_TAGS
    files = store_base.parse_files(deliverables, pipeline_tags, analysis_type_tags)

    return files


def _get_microbial_name(deliverables: dict) -> str:
    """ Get microbial id from deliverables """
    for file_ in deliverables["files"]:
        if file_["tag"] == "sampleinfo":
            return file_["id"]
    return None


def _get_date_from_deliverables_path(deliverables_path: str) -> dt.date:
    """ Get date from deliverables path """
    return dt.datetime.fromtimestamp(int(os.path.getctime(deliverables_path)))



"""cg module for cleaning databases and files"""
import logging
from datetime import datetime
from pathlib import Path
import shutil
import ruamel.yaml
import click
from dateutil.parser import parse as parse_date

from cg.apps.balsamic.fastq import FastqHandler
from cg.meta.workflow.balsamic import BalsamicAnalysisAPI
from cg.apps.hk import HousekeeperAPI
from cg.apps.lims import LimsAPI
from cg.apps.tb import TrailblazerAPI
from cg.apps.scoutapi import ScoutAPI
from cg.apps.crunchy import CrunchyAPI
from cg.apps.balsamic.api import BalsamicAPI
from cg.store import Store
from cg.constants import EXIT_SUCCESS, EXIT_FAIL

LOG = logging.getLogger(__name__)


@click.group()
@click.pass_context
def clean(context):
    """Clean up processes"""
    context.obj["store_api"] = Store(context.obj["database"])
    context.obj["hk_api"] = HousekeeperAPI(context.obj)
    context.obj["tb_api"] = TrailblazerAPI(context.obj)
    context.obj["scout_api"] = ScoutAPI(context.obj)
    context.obj["crunchy_api"] = CrunchyAPI(context.obj)
    context.obj["BalsamicAnalysisAPI"] = BalsamicAnalysisAPI(
        balsamic_api=BalsamicAPI(context.obj),
        store=context.obj["store_api"],
        housekeeper_api=context.obj["hk_api"],
        fastq_handler=FastqHandler(context.obj),
        lims_api=LimsAPI(context.obj),
    )


@clean.command("balsamic-run-dir")
@click.option("-y", "--yes", is_flag=True, help="Skip confirmation")
@click.option("-d", "--dry-run", is_flag=True, help="Shows cases and files that would be cleaned")
@click.argument("case_id")
@click.pass_context
def balsamic_run_dir(context, yes, case_id, dry_run: bool = False):
    """Remove Balsamic run directory"""

    balsamic_analysis_api = context.obj["BalsamicAnalysisAPI"]
    case_object = balsamic_analysis_api.get_case_object(case_id)
    if not case_object:
        LOG.warning(f"{case_id} not found!")
        raise click.Abort()

    analysis_obj = case_object.analyses[0] if case_object.analyses else None
    if not analysis_obj:
        LOG.error("%s: analysis not found", case_id)
        raise click.Abort()

    analysis_path = Path(balsamic_analysis_api.get_case_path(case_id))
    if yes or click.confirm(f"Do you want to remove {analysis_path}?"):
        if not analysis_path.exists():
            LOG.warning("could not find: %s", analysis_path)
            return EXIT_FAIL
        if analysis_path.is_symlink():
            LOG.warning(
                "Will not automatically delete symlink: %s, delete it manually", analysis_path
            )
            return EXIT_FAIL
        if dry_run:
            LOG.info("Would have deleted: %s", analysis_path)
            return EXIT_SUCCESS
        try:
            shutil.rmtree(analysis_path, ignore_errors=True)
            LOG.info(f"Cleaned {analysis_path}")
        except Exception as e:
            LOG.warning(
                f" Directory {analysis_path} will not be deleted due to unexpected error - {e}!"
            )

        analysis_obj.cleaned_at = datetime.now()
        balsamic_analysis_api.store.commit()


@clean.command("mip-run-dir")
@click.option("-y", "--yes", is_flag=True, help="Skip confirmation")
@click.option("-d", "--dry-run", is_flag=True, help="Shows cases and files that would be cleaned")
@click.argument("case_id")
@click.argument("sample_info", type=click.File("r"))
@click.pass_context
def mip_run_dir(context, yes, case_id, sample_info, dry_run: bool = False):
    """Remove MIP run directory"""

    raw_data = ruamel.yaml.safe_load(sample_info)
    date = context.obj["tb_api"].get_sampleinfo_date(raw_data)
    case_obj = context.obj["store_api"].family(case_id)

    if case_obj is None:
        LOG.error("%s: family not found", case_id)
        context.abort()

    analysis_obj = context.obj["store_api"].analysis(case_obj, date)
    if analysis_obj is None:
        LOG.error("%s - %s: analysis not found", case_id, date)
        context.abort()

    try:
        context.obj["tb_api"].delete_analysis(case_id, date, yes=yes, dry_run=dry_run)
        analysis_obj.cleaned_at = datetime.now()
    except ValueError as error:
        LOG.error(f"{case_id}: {error.args[0]}")
        context.abort()


@clean.command("hk-alignment-files")
@click.argument("bundle")
@click.option("-y", "--yes", is_flag=True, help="Skip confirmation")
@click.option("-d", "--dry-run", is_flag=True, help="Show files that would be cleaned")
@click.pass_context
def hk_alignment_files(context, bundle, yes: bool = False, dry_run: bool = False):
    """Clean up alignment files in Housekeeper bundle"""
    files = []
    for tag in ["bam", "bai", "bam-index", "cram", "crai", "cram-index"]:
        files.extend(context.obj["hk_api"].get_files(bundle=bundle, tags=[tag]))
    for file_obj in files:
        if file_obj.is_included:
            question = f"{bundle}: remove file from file system and database: {file_obj.full_path}"
        else:
            question = f"{bundle}: remove file from database: {file_obj.full_path}"

        if yes or click.confirm(question):
            file_name = file_obj.full_path
            if file_obj.is_included and Path(file_name).exists():
                if not dry_run:
                    Path(file_name).unlink()

            if not dry_run:
                file_obj.delete()
                context.obj["hk_api"].commit()
                click.echo(f"{file_name} deleted")


@clean.command("scout-finished-cases")
@click.option(
    "--days-old",
    type=int,
    default=300,
    help="Clean alignment files with analysis dates older then given number of days",
)
@click.option("-y", "--yes", is_flag=True, help="Skip confirmation")
@click.option("-d", "--dry-run", is_flag=True, help="Shows cases and files that would be cleaned")
@click.pass_context
def scout_finished_cases(context, days_old: int, yes: bool = False, dry_run: bool = False):
    """Clean up of solved and archived scout cases"""
    bundles = []
    for status in "archived", "solved":
        cases = context.obj["scout_api"].get_cases(status=status, reruns=False)
        cases_added = 0
        for case in cases:
            x_days_ago = datetime.now() - case.get("analysis_date")
            if x_days_ago.days > days_old:
                bundles.append(case.get("_id"))
                cases_added += 1
        LOG.info("%s cases marked for bam removal :)", cases_added)

    for bundle in bundles:
        context.invoke(hk_alignment_files, bundle=bundle, yes=yes, dry_run=dry_run)


@clean.command("hk-past-files")
@click.option("-c", "--case-id", type=str)
@click.option("-t", "--tags", multiple=True)
@click.option("-y", "--yes", is_flag=True, help="skip checks")
@click.option("-d", "--dry-run", is_flag=True, help="Shows cases and files that would be cleaned")
@click.pass_context
def hk_past_files(context, case_id, tags, yes, dry_run):
    """ Remove files found in older housekeeper bundles """
    if case_id:
        cases = [context.obj["store_api"].family(case_id)]
    else:
        cases = context.obj["store_api"].families()
    for case in cases:
        case_id = case.internal_id
        last_version = context.obj["hk_api"].last_version(bundle=case_id)
        if not last_version:
            continue
        last_version_file_paths = [
            Path(hk_file.full_path)
            for hk_file in context.obj["hk_api"].get_files(
                bundle=case_id, tags=None, version=last_version.id
            )
        ]
        LOG.info("Searching %s bundle for outdated files", case_id)
        hk_files = []
        for tag in tags:
            hk_files.extend(context.obj["hk_api"].get_files(bundle=case_id, tags=[tag]))
        for hk_file in hk_files:
            file_path = Path(hk_file.full_path)
            if file_path in last_version_file_paths:
                continue
            LOG.info("Will remove %s", file_path)
            if yes or click.confirm("Do you want to remove this file?"):
                if not dry_run:
                    hk_file.delete()
                    context.obj["hk_api"].commit()
                    if file_path.exists():
                        file_path.unlink()
                    LOG.info("File removed")


@clean.command("balsamic-past-run-dirs")
@click.option("-y", "--yes", is_flag=True, help="Skip confirmation")
@click.option("-d", "--dry-run", is_flag=True, help="Shows cases and files that would be cleaned")
@click.argument("before_str")
@click.pass_context
def balsamic_past_run_dirs(context, before_str: str, yes: bool = False, dry_run: bool = False):
    """Clean up of "old" Balsamic case run dirs"""

    before = parse_date(before_str)
    balsamic_analysis_api = context.obj["BalsamicAnalysisAPI"]
    possible_cleanups = balsamic_analysis_api.get_analyses_to_clean(before_date=before)
    LOG.info(f"Cleaning all analyses created before {before}")

    # for all analyses
    for analysis in possible_cleanups:
        case_id = analysis.family.internal_id

        # call clean
        LOG.info(f"Cleaning Balsamic output for {case_id}")
        try:
            context.invoke(balsamic_run_dir, yes=yes, case_id=case_id, dry_run=dry_run)
        except click.Abort:
            continue

    LOG.info("Done cleaning Balsamic output")


@clean.command("mip-past-run-dirs")
@click.option("-y", "--yes", is_flag=True, help="Skip confirmation")
@click.option("-d", "--dry-run", is_flag=True, help="Shows cases and files that would be cleaned")
@click.argument("before_str")
@click.pass_context
def mip_past_run_dirs(
    context: click.Context, before_str: str, yes: bool = False, dry_run: bool = False
):
    """Clean up of "old" MIP case run dirs"""
    before = parse_date(before_str)
    old_analyses = context.obj["store_api"].analyses(before=before)
    for status_analysis in old_analyses:
        case_id = status_analysis.family.internal_id
        LOG.debug("%s: clean up analysis output", case_id)
        tb_analysis = context.obj["tb_api"].find_analysis(
            family=case_id, started_at=status_analysis.started_at, status="completed"
        )

        if tb_analysis is None:
            LOG.warning("%s: analysis not found in Trailblazer", case_id)
            continue
        elif tb_analysis.is_deleted:
            LOG.warning("%s: analysis already deleted", case_id)
            continue
        elif context.obj["tb_api"].analyses(family=case_id, temp=True).count() > 0:
            LOG.warning("%s: family already re-started", case_id)
            continue

        try:
            sampleinfo_path = context.obj["tb_api"].get_sampleinfo(tb_analysis)
            LOG.info("%s: cleaning MIP output", case_id)
            with open(sampleinfo_path, "r") as sampleinfo_file:
                context.invoke(
                    mip_run_dir,
                    yes=yes,
                    case_id=case_id,
                    sample_info=sampleinfo_file,
                    dry_run=dry_run,
                )
        except FileNotFoundError:
            LOG.error(
                (
                    "%s: sample_info file not found, please mark the analysis as deleted in the "
                    "analysis table in trailblazer."
                ),
                case_id,
            )



"""Click commands to store microSALT analyses"""
import logging

import click

from cg.apps import hk, tb
from cg.exc import (
    AnalysisDuplicationError,
    BundleAlreadyAddedError,
    MandatoryFilesMissing,
)
from cg.meta.store.base import gather_files_and_bundle_in_housekeeper
from cg.store import Store

from cg.constants import EXIT_SUCCESS, EXIT_FAIL

import _io

LOG = logging.getLogger(__name__)


@click.group()
@click.pass_context
def store(context):
    """Store results from microSALT in housekeeper."""
    context.obj["db"] = Store(context.obj["database"])
    context.obj["tb_api"] = tb.TrailblazerAPI(context.obj)
    context.obj["hk_api"] = hk.HousekeeperAPI(context.obj)


@store.command()
@click.argument("config-stream", type=click.File("r"), required=False)
@click.pass_context
def analysis(context, config_stream):
    """Store a finished analysis in Housekeeper."""
    status = context.obj["db"]
    hk_api = context.obj["hk_api"]

    if not config_stream:
        LOG.error("Please provide a config file")
        context.abort()

    try:
        new_analysis = gather_files_and_bundle_in_housekeeper(
            config_stream, hk_api, status, workflow="microsalt"
        )
    except (AnalysisDuplicationError, BundleAlreadyAddedError, MandatoryFilesMissing) as error:
        click.echo(click.style(error.message, fg="red"))
        context.abort()
    except FileNotFoundError as error:
        click.echo(click.style(f"missing file: {error.args[0]}", fg="red"))
        context.abort()

    status.add_commit(new_analysis)
    click.echo(click.style("included files in Housekeeper", fg="green"))


@store.command()
@click.pass_context
def completed(context):
    """Store all completed analyses"""
    microsalt_analysis_api = context.obj["analysis_api"]
    exit_code = EXIT_SUCCESS
    for deliverables_file in microsalt_analysis_api.get_deliverables_to_store():
        try:
            context.invoke(analysis, config_stream=_io.open(deliverables_file))
        except click.Abort:
            exit_code = EXIT_FAIL
        except Exception as error:
            LOG.error("Unspecified error occurred - %s", error)
            exit_code = EXIT_FAIL
    if exit_code:
        raise click.Abort



from typing import List
import logging

from genologics.entities import Project, Researcher, Sample, Container, Containertype, Artifact
from lxml import etree
from lxml.objectify import ObjectifiedElement

from cg.exc import OrderError
from .constants import PROP2UDF
from . import batch

LOG = logging.getLogger(__name__)
CONTAINER_TYPE_MAP = {"Tube": 2, "96 well plate": 1}


class OrderHandler:
    def save_xml(self, uri: str, document: ObjectifiedElement):
        """Post the data to the server."""
        data = etree.tostring(document, xml_declaration=True)
        LOG.debug(data)
        result = self.post(uri, data)
        return result

    def save_containers(self, container_details: ObjectifiedElement):
        """Save a batch of containers."""
        container_uri = f"{self.get_uri()}/containers/batch/create"
        results = self.save_xml(container_uri, container_details)
        container_map = {}
        for link in results.findall("link"):
            lims_container = Container(self, uri=link.attrib["uri"])
            container_map[lims_container.name] = lims_container
        return container_map

    def save_samples(self, sample_details: ObjectifiedElement, map_samples=False):
        """Save a batch of samples."""
        sample_uri = f"{self.get_uri()}/samples/batch/create"
        results = self.save_xml(sample_uri, sample_details)
        if map_samples:
            sample_map = {}
            for link in results.findall("link"):
                lims_sample = Sample(self, uri=link.attrib["uri"])
                sample_map[lims_sample.name] = lims_sample
            return sample_map
        return results

    def update_artifacts(self, artifact_details: ObjectifiedElement):
        """Update/put a batch of artifacts."""
        artifact_uri = f"{self.get_uri()}/artifacts/batch/update"
        results = self.save_xml(artifact_uri, artifact_details)
        return results

    def submit_project(self, project_name: str, samples: List[dict], researcher_id: str = "3"):
        """Parse Scout project."""
        containers = self.prepare(samples)

        lims_project = Project.create(
            self, researcher=Researcher(self, id=researcher_id), name=project_name
        )
        LOG.info("%s: created new LIMS project", lims_project.id)

        containers_data = [
            batch.build_container(
                name=container["name"], con_type=Containertype(lims=self, id=container["type"])
            )
            for container in containers
        ]
        container_details = batch.build_container_batch(containers_data)
        LOG.debug("%s: saving containers", lims_project.name)
        container_map = self.save_containers(container_details)

        reagentlabel_samples = [
            sample
            for container in containers
            for sample in container["samples"]
            if sample["index_sequence"]
        ]

        samples_data = []
        for container in containers:
            for sample in container["samples"]:
                LOG.debug("%s: adding sample to container: %s", sample["name"], container["name"])
                lims_container = container_map[container["name"]]
                sample_data = batch.build_sample(
                    name=sample["name"],
                    project=lims_project,
                    container=lims_container,
                    location=sample["location"],
                    udfs=sample["udfs"],
                )
                samples_data.append(sample_data)
        sample_details = batch.build_sample_batch(samples_data)
        process_reagentlabels = len(reagentlabel_samples) > 0
        sample_map = self.save_samples(sample_details, map_samples=process_reagentlabels)

        if process_reagentlabels:
            artifacts_data = [
                batch.build_artifact(
                    artifact=sample_map[sample["name"]].artifact,
                    reagent_label=sample["index_sequence"],
                )
                for sample in reagentlabel_samples
            ]
            artifact_details = batch.build_artifact_batch(artifacts_data)
            self.update_artifacts(artifact_details)

        lims_project_data = self._export_project(lims_project)
        return lims_project_data

    @classmethod
    def prepare(cls, samples):
        """Convert API input to LIMS input data."""
        lims_containers = []
        tubes, plates = cls.group_containers(samples)
        # "96 well plate" = container type "1"; Tube = container type "2"
        for container_type, containers in [("1", plates), ("2", tubes)]:
            for container_name, samples in containers.items():
                new_container = {"name": container_name, "type": container_type, "samples": []}
                # check that positions in plate are unique
                well_positions = {}
                for sample_data in samples:
                    location = sample_data["well_position"] or None
                    if location:
                        if location in well_positions:
                            first_sample = well_positions[location]
                            message = (
                                f"duplicate well position: {location} | {first_sample}"
                                f" - {sample_data['name']}"
                            )
                            raise OrderError(message)
                        well_positions[location] = sample_data["name"]
                    if sample_data["container"] == "96 well plate" and location is None:
                        message = f"missing 'well_position' for sample: {sample_data['name']}"
                        raise ValueError(message)
                    new_sample = {
                        "name": sample_data["name"],
                        "location": location or "1:1",
                        "index_sequence": sample_data["index_sequence"],
                        "udfs": {},
                    }
                    for key, value in sample_data["udfs"].items():
                        if value is None:
                            LOG.debug(f"{key}: skipping null value UDF")
                            continue
                        if key in PROP2UDF:
                            if isinstance(value, bool):
                                value = "yes" if value else "no"
                            new_sample["udfs"][PROP2UDF[key]] = value
                        else:
                            LOG.debug(f"UDF not found: {key} - {value}")

                    new_container["samples"].append(new_sample)
                lims_containers.append(new_container)
        return lims_containers

    @staticmethod
    def group_containers(samples):
        """Group samples by containers."""
        tubes = {}
        plates = {}
        for sample_data in samples:
            if sample_data["container"] == "Tube":
                # detected tube: name after sample unless specified
                container_name = sample_data.get("container_name") or sample_data["name"]
                if container_name in tubes:
                    raise OrderError(f"{container_name}: conflicting sample/tube name")
                tubes[container_name] = [sample_data]
            elif sample_data["container"] == "96 well plate":
                # detected plate: require container name
                if sample_data["container_name"] not in plates:
                    plates[sample_data["container_name"]] = []
                plates[sample_data["container_name"]].append(sample_data)
            else:
                raise ValueError(f"unknown container type: {sample_data['container']}")
        return tubes, plates



"""Tests for the compress fastq cli"""

import logging

from cg.cli.store import fastq_cmd


def test_store_fastq_cli_no_family(compress_context, cli_runner, caplog):
    """Test to run the compress command without providing any case id"""
    caplog.set_level(logging.DEBUG)
    # GIVEN a context

    # WHEN running the store fastq command
    res = cli_runner.invoke(fastq_cmd, [], obj=compress_context)

    # THEN assert the program exits with 2 since no argument was provided
    assert res.exit_code == 2


def test_store_fastq_cli_non_existing_family(compress_context, cli_runner, caplog):
    """Test to run the compress command witha non existing case"""
    caplog.set_level(logging.DEBUG)
    # GIVEN a context and a non existing case id
    case_id = "happychap"

    # WHEN running the store fastq command
    res = cli_runner.invoke(fastq_cmd, [case_id], obj=compress_context)

    # THEN assert the program exits without a problem
    assert res.exit_code == 0
    # THEN assert it was communicated that no families where found
    assert f"Could not find case {case_id}" in caplog.text



import json

import pytest
from cg.apps.lims import LimsAPI

from cg.apps.osticket import OsTicket
from cg.meta.orders import OrdersAPI, OrderType
from cg.meta.orders.status import StatusHandler


class MockLims(LimsAPI):

    lims = None

    def __init__(self):
        self.lims = self

    def update_sample(
        self,
        lims_id: str,
        sex=None,
        application: str = None,
        target_reads: int = None,
        priority=None,
    ):
        pass


@pytest.fixture
def mip_order_to_submit():
    """Load an example scout order."""
    return json.load(open("tests/fixtures/orders/mip.json"))


@pytest.fixture
def mip_rna_order_to_submit():
    """Load an example rna order."""
    return json.load(open("tests/fixtures/orders/mip_rna.json"))


@pytest.fixture
def external_order_to_submit():
    """Load an example external order."""
    return json.load(open("tests/fixtures/orders/external.json"))


@pytest.fixture
def fastq_order_to_submit():
    """Load an example fastq order."""
    return json.load(open("tests/fixtures/orders/fastq.json"))


@pytest.fixture
def rml_order_to_submit():
    """Load an example rml order."""
    return json.load(open("tests/fixtures/orders/rml.json"))


@pytest.fixture
def metagenome_order_to_submit():
    """Load an example metagenome order."""
    return json.load(open("tests/fixtures/orders/metagenome.json"))


@pytest.fixture
def microbial_order_to_submit():
    """Load an example microbial order."""
    return json.load(open("tests/fixtures/orders/microbial.json"))


@pytest.fixture
def balsamic_order_to_submit():
    """Load an example cancer order."""
    return json.load(open("tests/fixtures/orders/balsamic.json"))


@pytest.fixture
def mip_balsamic_order_to_submit():
    """Load an example scout order."""
    return json.load(open("tests/fixtures/orders/mip_balsamic.json"))


@pytest.fixture
def all_orders_to_submit(
    rml_order_to_submit,
    fastq_order_to_submit,
    mip_order_to_submit,
    mip_rna_order_to_submit,
    external_order_to_submit,
    microbial_order_to_submit,
    metagenome_order_to_submit,
    balsamic_order_to_submit,
    mip_balsamic_order_to_submit,
):
    return {
        OrderType.RML: rml_order_to_submit,
        OrderType.FASTQ: fastq_order_to_submit,
        OrderType.MIP: mip_order_to_submit,
        OrderType.MIP_RNA: mip_rna_order_to_submit,
        OrderType.EXTERNAL: external_order_to_submit,
        OrderType.MICROBIAL: microbial_order_to_submit,
        OrderType.METAGENOME: metagenome_order_to_submit,
        OrderType.BALSAMIC: balsamic_order_to_submit,
        OrderType.MIP_BALSAMIC: mip_balsamic_order_to_submit,
    }


@pytest.fixture
def rml_status_data(rml_order_to_submit):
    """Parse rml order example."""
    data = StatusHandler.pools_to_status(rml_order_to_submit)
    return data


@pytest.fixture
def fastq_status_data(fastq_order_to_submit):
    """Parse fastq order example."""
    data = StatusHandler.samples_to_status(fastq_order_to_submit)
    return data


@pytest.fixture
def mip_status_data(mip_order_to_submit):
    """Parse scout order example."""
    data = StatusHandler.cases_to_status(mip_order_to_submit)
    return data


@pytest.fixture
def mip_rna_status_data(mip_rna_order_to_submit):
    """Parse rna order example."""
    data = StatusHandler.cases_to_status(mip_rna_order_to_submit)
    return data


@pytest.fixture
def external_status_data(external_order_to_submit):
    """Parse external order example."""
    data = StatusHandler.cases_to_status(external_order_to_submit)
    return data


@pytest.fixture
def microbial_status_data(microbial_order_to_submit):
    """Parse microbial order example."""
    data = StatusHandler.microbial_samples_to_status(microbial_order_to_submit)
    return data


@pytest.fixture
def metagenome_status_data(metagenome_order_to_submit):
    """Parse metagenome order example."""
    data = StatusHandler.samples_to_status(metagenome_order_to_submit)
    return data


@pytest.fixture
def balsamic_status_data(balsamic_order_to_submit):
    """Parse cancer order example."""
    data = StatusHandler.cases_to_status(balsamic_order_to_submit)
    return data


@pytest.fixture(scope="function")
def orders_api(base_store):
    osticket_api = OsTicket()
    lims = MockLims()
    _orders_api = OrdersAPI(lims=lims, status=base_store, osticket=osticket_api)
    return _orders_api



"""Fixtures for compress api tests"""
import copy
import json
import os
from pathlib import Path

import pytest

from cg.apps.crunchy import CrunchyAPI
from cg.meta.compress import CompressAPI


class CompressionData:
    """Class to hold some data files"""

    def __init__(self, **kwargs):
        self.spring_path = kwargs["spring_path"]
        self.pending_path = kwargs["pending_path"]
        self.spring_metadata_path = kwargs["spring_metadata_path"]
        self.fastq_first = kwargs["fastq_first"]
        self.fastq_second = kwargs["fastq_second"]

    @property
    def spring_file(self):
        """Return the path to an existing spring file"""
        self.spring_path.touch()
        return self.spring_path

    @property
    def pending_file(self):
        """Return the path to an existing pending file"""
        self.pending_path.touch()
        return self.pending_path

    @staticmethod
    def _spring_metadata(fastq_first, fastq_second, spring_path, updated=False, date=None):
        """Return metada information"""
        metadata = [
            {
                "path": str(fastq_first.resolve()),
                "file": "first_read",
                "checksum": "checksum_first_read",
                "algorithm": "sha256",
            },
            {
                "path": str(fastq_second.resolve()),
                "file": "second_read",
                "checksum": "checksum_second_read",
                "algorithm": "sha256",
            },
            {"path": str(spring_path.resolve()), "file": "spring"},
        ]
        if updated:
            for file_info in metadata:
                file_info["updated"] = date or "2020-01-21"
        return metadata

    @property
    def spring_metadata_file(self):
        """Return the path to an existing spring metadata file"""

        spring_metadata = CompressionData._spring_metadata(
            self.fastq_first, self.fastq_second, self.spring_path
        )
        with open(self.spring_metadata_path, "w") as outfile:
            outfile.write(json.dumps(spring_metadata))

        return self.spring_metadata_path

    @property
    def updated_spring_metadata_file(self):
        """Return the path to an existing spring metadata file"""

        spring_metadata = CompressionData._spring_metadata(
            self.fastq_first, self.fastq_second, self.spring_path, True
        )
        with open(self.spring_metadata_path, "w") as outfile:
            outfile.write(json.dumps(spring_metadata))

        return self.spring_metadata_path

    @property
    def fastq_first_file(self):
        """Return the path to an existing spring metadata file"""
        self.fastq_first.touch()
        return self.fastq_first

    @property
    def fastq_second_file(self):
        """Return the path to an existing fastq file"""
        self.fastq_second.touch()
        return self.fastq_second


@pytest.yield_fixture(scope="function", name="compression_files")
def fixture_compression_files(compression_object):
    """Return a small class with compression files"""
    return CompressionData(
        spring_path=compression_object.spring_path,
        spring_metadata_path=compression_object.spring_metadata_path,
        fastq_first=compression_object.fastq_first,
        fastq_second=compression_object.fastq_second,
        pending_path=compression_object.pending_path,
    )


@pytest.yield_fixture(scope="function", name="real_crunchy_api")
def fixture_real_crunchy_api(crunchy_config_dict):
    """crunchy api fixture"""

    yield CrunchyAPI(crunchy_config_dict)


@pytest.yield_fixture(scope="function", name="compress_api")
def fixture_compress_api(real_crunchy_api, housekeeper_api):
    """compress api fixture"""
    hk_api = housekeeper_api
    _api = CompressAPI(crunchy_api=real_crunchy_api, hk_api=hk_api)
    yield _api


@pytest.fixture(scope="function", name="populated_compress_fastq_api")
def fixture_populated_compress_fastq_api(compress_api, compress_hk_fastq_bundle, helpers):
    """Populated compress api fixture"""
    helpers.ensure_hk_bundle(compress_api.hk_api, compress_hk_fastq_bundle)

    return compress_api


@pytest.fixture(scope="function", name="populated_decompress_spring_api")
def fixture_populated_decompress_spring_api(compress_api, decompress_hk_spring_bundle, helpers):
    """Populated compress api fixture"""
    helpers.ensure_hk_bundle(compress_api.hk_api, decompress_hk_spring_bundle)

    return compress_api


@pytest.fixture(scope="function", name="sample")
def fixture_sample():
    """Return the sample id for first sample"""
    return "sample_1"


@pytest.fixture(scope="function", name="spring_path")
def fixture_spring_path(compression_object) -> Path:
    """Return the path to a non existing spring file"""
    return compression_object.spring_path


@pytest.fixture(scope="function", name="spring_metadata_path")
def fixture_spring_metadata_path(compression_object) -> Path:
    """Return the path to a non existing spring metadata file"""
    return compression_object.spring_metadata_path


@pytest.fixture(scope="function", name="fastq_flag_path")
def fixture_fastq_flag_path(spring_metadata_path) -> Path:
    """Return the path to a non existing fastq flag file"""
    return spring_metadata_path


@pytest.fixture(scope="function", name="fastq_flag_file")
def fixture_fastq_flag_file(spring_metadata_path) -> Path:
    """Return the path to an existing fastq flag file"""
    spring_metadata_path.touch()
    return spring_metadata_path


@pytest.fixture(scope="function", name="spring_file")
def fixture_spring_file(spring_path) -> Path:
    """Return the path to an existing spring file"""
    spring_path.touch()
    return spring_path


@pytest.fixture(scope="function", name="multi_linked_file")
def fixture_multi_linked_file(spring_path, project_dir) -> Path:
    """Return the path to an existing file with two links"""
    first_link = project_dir / "link-1"
    os.link(spring_path, first_link)

    return spring_path


@pytest.fixture(scope="function", name="fastq_paths")
def fixture_fastq_paths(compression_object):
    """Fixture for temporary fastq-files

    Create fastq paths and return a dictionary with them
    """
    return {
        "fastq_first_path": compression_object.fastq_first,
        "fastq_second_path": compression_object.fastq_second,
    }


@pytest.fixture(scope="function", name="fastq_files")
def fixture_fastq_files(fastq_paths):
    """Fixture for temporary fastq-files

    Create fastq files and return a dictionary with them
    """
    fastq_first_file = fastq_paths["fastq_first_path"]
    fastq_second_file = fastq_paths["fastq_second_path"]
    fastq_first_file.touch()
    fastq_second_file.touch()

    return fastq_paths


# Bundle fixtures


@pytest.fixture(scope="function", name="decompress_hk_spring_bundle")
def fixture_decompress_hk_spring_bundle(
    sample_hk_bundle_no_files, spring_file, fastq_flag_file, sample
):
    """Create a complete bundle mock for testing decompression

    This bundle contains a spring file and a spring metadata file
    """
    hk_bundle_data = copy.deepcopy(sample_hk_bundle_no_files)

    spring_file_info = {"path": str(spring_file), "archive": False, "tags": [sample, "spring"]}
    spring_meta_info = {
        "path": str(fastq_flag_file),
        "archive": False,
        "tags": [sample, "spring-metadata"],
    }

    hk_bundle_data["files"].append(spring_file_info)
    hk_bundle_data["files"].append(spring_meta_info)

    return hk_bundle_data



""" Common MIP related functionality """

import logging

from cg.constants import SPACE
from cg.utils import Process

LOG = logging.getLogger(__name__)

"""
This dict is built like this:
'cg-option': {
     'option': '--mip-option
}

Defaults can also be set:
'cg-option': {
    'option': '--mip-option',
    'default': '1'
}

This dict is used in `build_command`.

"""
CLI_OPTIONS = {
    "config": {"option": "--config_file"},
    "priority": {"option": "--slurm_quality_of_service"},
    "email": {"option": "--email"},
    "base": {"option": "--cluster_constant_path"},
    "dryrun": {"option": "--dry_run_all"},
    "gene_list": {"option": "--vcfparser_slt_fl"},
    "max_gaussian": {"option": "--gatk_varrecal_snv_max_gau"},
    "skip_evaluation": {"option": "--qccollect_skip_evaluation"},
    "start_with": {"option": "--start_with_recipe"},
}


class MipAPI:
    """ Group MIP specific functionality """

    def __init__(self, script: str, pipeline: str, conda_env: str, **kwargs):
        """Initialize MIP command line interface"""
        self.script = script
        self.pipeline = pipeline
        self.conda_env = conda_env

    def run_command(self, config: str, case: str, dry_run: bool = False, **kwargs) -> None:
        """Execute the workflow"""
        command = self.build_command(config, case, **kwargs)
        self.execute(command, dry_run)

    def build_command(self, config: str, case: str, **kwargs) -> dict:
        """Build the command dict to execute MIP"""
        binary_command = [self.script, self.pipeline, case]
        parameters = []
        for key, value in kwargs.items():
            if value:
                _cg_to_mip_option_map(parameters, key)
                _append_value_for_non_flags(parameters, value)

        command = {
            "binary": SPACE.join(binary_command),
            "config": config,
            "environment": self.conda_env,
            "parameters": parameters,
        }
        return command

    @classmethod
    def execute(cls, command: dict, dry_run: bool = False) -> int:
        """Start a new MIP analysis
        Args:
            command(dict): Command to execute
            dry_run(bool): Print command instead of executing it
        """
        mip_process = Process(
            command["binary"], config=command["config"], environment=command["environment"]
        )

        process_return_code = mip_process.run_command(
            dry_run=dry_run, parameters=command["parameters"]
        )
        for line in mip_process.stdout_lines():
            LOG.info(line)
        for line in mip_process.stderr_lines():
            LOG.info(line)
        return process_return_code


def _append_value_for_non_flags(parameters: list, value):
    """Add the value of the non boolean options to the parameters"""
    if value is not True:
        parameters.append(value)


def _cg_to_mip_option_map(parameters: list, mip_key):
    """Map cg options to MIP option syntax"""
    parameters.append(CLI_OPTIONS[mip_key]["option"])



""" Module to decouple cg code from Housekeeper code """
import datetime as dt
import logging
import os
from pathlib import Path
from typing import List, Tuple

from housekeeper.include import checksum as hk_checksum
from housekeeper.include import include_version
from housekeeper.store import Store, models

LOG = logging.getLogger(__name__)


class HousekeeperAPI:
    """ API to decouple cg code from Housekeeper """

    def __init__(self, config: dict) -> None:
        self._store = Store(config["housekeeper"]["database"], config["housekeeper"]["root"])
        self.root_dir = config["housekeeper"]["root"]

    def __getattr__(self, name):
        LOG.warning("Called undefined %s on %s, please wrap", name, self.__class__.__name__)
        return getattr(self._store, name)

    def new_bundle(self, name: str, created_at: dt.datetime = None) -> models.Bundle:
        """ Create a new file bundle """
        return self._store.new_bundle(name, created_at)

    def add_bundle(self, bundle_data) -> Tuple[models.Bundle, models.Version]:
        """ Build a new bundle version of files """
        return self._store.add_bundle(bundle_data)

    def bundle(self, name: str) -> models.Bundle:
        """ Fetch a bundle """
        return self._store.bundle(name)

    def bundles(self) -> List[models.Bundle]:
        """ Fetch bundles """
        return self._store.bundles()

    def new_file(
        self, path: str, checksum: str = None, to_archive: bool = False, tags: list = None
    ) -> models.File:
        """ Create a new file """
        if tags is None:
            tags = []
        return self._store.new_file(path, checksum, to_archive, tags)

    def add_file(self, path, version_obj: models.Version, tags, to_archive=False) -> models.File:
        """Add a file to the database"""
        if isinstance(tags, str):
            tags = [tags]
        for tag_name in tags:
            if not self.tag(tag_name):
                self.add_tag(tag_name)

        new_file = self.new_file(
            path=str(Path(path).absolute()),
            to_archive=to_archive,
            tags=[self.tag(tag_name) for tag_name in tags],
        )

        new_file.version = version_obj
        return new_file

    def files(
        self, *, bundle: str = None, tags: List[str] = None, version: int = None, path: str = None
    ) -> List[models.File]:
        """ Fetch files """
        return self._store.files(bundle=bundle, tags=tags, version=version, path=path)

    def rollback(self):
        """ Wrap method in Housekeeper Store """
        return self._store.rollback()

    def session_no_autoflush(self):
        """ Wrap property in Housekeeper Store """
        return self._store.session.no_autoflush

    def get_files(self, bundle: str, tags: list, version: int = None):
        """Fetch all the files in housekeeper, optionally filtered by bundle and/or tags and/or
        version

        Returns:
            iterable(hk.Models.File)
        """
        return self._store.files(bundle=bundle, tags=tags, version=version)

    @staticmethod
    def get_included_path(
        root_dir: Path, version_obj: models.Version, file_obj: models.File
    ) -> Path:
        """Generate the path to a file that should be included.
        If the version dir does not exist, create a new version dir in root dir
        """
        # generate root directory
        version_root_dir = root_dir / version_obj.relative_root_dir
        version_root_dir.mkdir(parents=True, exist_ok=True)
        LOG.info("Created new bundle version dir: %s", version_root_dir)
        return version_root_dir / Path(file_obj.path).name

    def include_file(self, file_obj: models.File, version_obj: models.Version) -> models.File:
        """Call the include version function to import related assets."""
        global_root_dir = Path(self.get_root_dir())

        new_path = self.get_included_path(global_root_dir, version_obj, file_obj)
        if file_obj.to_archive:
            # calculate sha1 checksum if file is to be archived
            file_obj.checksum = HousekeeperAPI.checksum(file_obj.path)
        # hardlink file to the internal structure
        os.link(file_obj.path, new_path)
        LOG.info("Linked file: %s -> %s", file_obj.path, new_path)
        file_obj.path = str(new_path).replace(f"{global_root_dir}/", "", 1)
        return file_obj

    def new_version(
        self, created_at: dt.datetime, expires_at: dt.datetime = None
    ) -> models.Version:
        """ Create a new bundle version """
        return self._store.new_version(created_at, expires_at)

    def version(self, bundle: str, date: dt.datetime) -> models.Version:
        """ Fetch a version """
        return self._store.version(bundle, date)

    def last_version(self, bundle: str) -> models.Version:
        """Gets the latest version of a bundle"""
        return (
            self._store.Version.query.join(models.Version.bundle)
            .filter(models.Bundle.name == bundle)
            .order_by(models.Version.created_at.desc())
            .first()
        )

    def new_tag(self, name: str, category: str = None):
        """ Create a new tag """
        return self._store.new_tag(name, category)

    def add_tag(self, name: str, category: str = None):
        """ Add a tag to the database """
        tag_obj = self._store.new_tag(name, category)
        self.add_commit(tag_obj)
        return tag_obj

    def tag(self, name: str):
        """ Fetch a tag """
        return self._store.tag(name)

    def include(self, version_obj: models.Version):
        """Call the include version function to import related assets."""
        include_version(self.get_root_dir(), version_obj)
        version_obj.included_at = dt.datetime.now()

    def add_commit(self, *args, **kwargs):
        """ Wrap method in Housekeeper Store """
        return self._store.add_commit(*args, **kwargs)

    def commit(self):
        """ Wrap method in Housekeeper Store """
        return self._store.commit()

    def session_no_autoflush(self):
        """ Wrap property in Housekeeper Store """
        return self._store.session.no_autoflush

    def get_root_dir(self):
        """Returns the root dir of Housekeeper"""
        return self.root_dir

    @staticmethod
    def checksum(path):
        """Calculate the checksum"""
        return hk_checksum(path)

    def initialise_db(self):
        """Create all tables in the store."""
        self._store.create_all()

    def destroy_db(self):
        """Drop all tables in the store"""
        self._store.drop_all()



class MockLimsAPI:
    """Mock LIMS API to get target bed from lims"""

    def __init__(self, config: dict = None):
        self.config = config
        self.sample_vars = {}

    def add_sample(self, internal_id: str):
        self.sample_vars[internal_id] = {}

    def add_capture_kit(self, internal_id: str, capture_kit):
        if not internal_id in self.sample_vars:
            self.add_sample(internal_id)
        self.sample_vars[internal_id]["capture_kit"] = capture_kit

    def capture_kit(self, internal_id: str):
        if internal_id in self.sample_vars:
            return self.sample_vars[internal_id].get("capture_kit")
        return None



from cg.store import Store


def test_add_customer(invoke_cli, disk_store: Store):
    # GIVEN an empty database

    # WHEN adding a customer
    db_uri = disk_store.uri
    result = invoke_cli(
        [
            "--database",
            db_uri,
            "add",
            "customer",
            "internal_id",
            "testcust",
            "--invoice-address",
            "Test adress",
            "--invoice-reference",
            "ABCDEF",
        ]
    )

    # THEN it should be stored in the database
    assert result.exit_code == 0
    assert disk_store.Customer.query.count() == 1


def test_add_user(invoke_cli, disk_store: Store):
    # GIVEN a database with a customer in it that we can connect the user to
    customer_id = "custtest"
    customer_group = disk_store.add_customer_group("dummy_group", "dummy group")
    customer = disk_store.add_customer(
        internal_id=customer_id,
        name="Test Customer",
        scout_access=False,
        customer_group=customer_group,
        invoice_address="Street nr, 12345 Uppsala",
        invoice_reference="ABCDEF",
    )
    disk_store.add_commit(customer)

    # WHEN adding a new user
    name, email = "Paul T. Anderson", "paul.anderson@magnolia.com"
    db_uri = disk_store.uri
    result = invoke_cli(["--database", db_uri, "add", "user", "-c", customer_id, email, name])

    # THEN it should be stored in the database
    assert result.exit_code == 0
    assert disk_store.User.query.count() == 1



"""Fixtures for cli balsamic tests"""

import pytest

from cg.apps.lims import LimsAPI
from cg.store import Store


@pytest.fixture
def base_context(base_store: Store) -> dict:
    """context to use in cli"""
    return {"lims": MockLims(), "status": base_store}


class MockLims(LimsAPI):
    """Mock lims fixture"""

    lims = None

    def __init__(self):
        self.lims = self

    _project_name = None
    _updated_key = None
    _updated_value = None

    def update_project(self, lims_id: str, name=None):
        """Mock lims update_project"""
        self._project_name = name

    def get_updated_project_name(self) -> str:
        """Method to help test that update project was called with name parameter"""
        return self._project_name

    def update_sample(self, lims_id: str, **kwargs):
        """Mock lims update_sample"""
        for key, value in kwargs.items():
            # n.b. only works when len(kwargs)==1
            self._updated_key = key
            self._updated_value = value

    def get_updated_sample_key(self) -> str:
        """Method to help test that update sample was called with key parameter"""
        return self._updated_key

    def get_updated_sample_value(self) -> str:
        """Method to help test that update sample was called with value"""
        return self._updated_value



"""Functions to get cases from the store database"""

from typing import Iterator

from sqlalchemy.orm import Query

from cg.store import models
from cg.store.api import Store


def analysis_completed(store: Store) -> Query:
    """Return cases where analysis is finished"""
    cases = store.Family.query.join(models.Analysis)
    return cases.filter(models.Analysis.completed_at.isnot(None)).order_by(
        models.Family.created_at.asc()
    )


def ready_for_spring_compresssion(store: Store) -> Iterator[models.Family]:
    """Fetch all cases that are ready to be compressed by SPRING"""
    for case_obj in analysis_completed(store):
        if case_obj.ordered_at > case_obj.analyses[0].completed_at:
            continue
        yield case_obj



"""
    Module to upload cases to mutacc
"""

import os
import logging
from collections import namedtuple

from cg.apps import scoutapi, mutacc_auto

LOG = logging.getLogger(__name__)


class UploadToMutaccAPI:

    """API to upload finished cases to mutacc"""

    def __init__(self, scout_api: scoutapi.ScoutAPI, mutacc_auto_api: mutacc_auto.MutaccAutoAPI):

        self.scout = scout_api
        self.mutacc_auto = mutacc_auto_api

    def extract_reads(self, case: dict):
        """Use mutacc API to extract reads from case"""
        data = self.data(case)
        if data:
            LOG.info("Extracting reads from case %s", case["_id"])
            self.mutacc_auto.extract_reads(case=data["case"], variants=data["causatives"])

    def import_cases(self):
        """Use mutacc API to import cases to database"""
        LOG.info("importing cases into mutacc database")
        self.mutacc_auto.import_reads()

    def data(self, case) -> dict:
        """
        Find the necessary data for the case

        Args:
            case (dict): case dictionary from scout

        Returns:
            data (dict): dictionary with case data, and data on causative variants
        """

        if all([self._has_bam(case), self._has_causatives(case)]):
            causatives = self.scout.get_causative_variants(case_id=case["_id"])
            mutacc_case = remap(case, SCOUT_TO_MUTACC_CASE)
            mutacc_variants = [remap(variant, SCOUT_TO_MUTACC_VARIANTS) for variant in causatives]
            return {"case": mutacc_case, "causatives": mutacc_variants}
        return {}

    @staticmethod
    def _has_bam(case: dict) -> bool:

        """
        Check that all samples in case has a given path to a bam file,
        and that the file exists

        Args:
            case (dict): case dictionary from scout

        Returns:
            (bool): True if all samples has valid paths to a bam-file

        """
        if case.get("individuals") is None:
            LOG.warning("case dictionary is missing 'individuals' key")
            raise KeyError

        for sample in case["individuals"]:

            if sample.get("bam_file") is None:
                LOG.info(
                    "sample %s in case %s is missing bam file. skipping",
                    sample["individual_id"],
                    case["_id"],
                )
                return False

            if not os.path.isfile(sample["bam_file"]):
                LOG.info(
                    "sample %s in %s has non existing bam file. skipping",
                    sample["individual_id"],
                    case["_id"],
                )
                return False

        return True

    @staticmethod
    def _has_causatives(case: dict) -> bool:
        """
        Check that the case has marked causative variants in scout

        Args:
            case (dict): case dictionary from scout

        Returns:
            (bool): True if case has marked causative variants in scout
        """
        if case.get("causatives"):
            return True

        LOG.info("case %s has no marked causatives in scout", case["_id"])
        return False


# Reformat scout noutput to mutacc input

MAPPER = namedtuple("mapper", ["field_name_1", "field_name_2", "conv"])


def remap(input_dict: dict, mapper_list: list) -> dict:
    """
    Reformat dict from one application to be used by another

    Args:
        input_dict (dict): dictionary to be converted
        mapper_list (list(MAPPER)): list of mapper objects

    Returns:
        output_dict (dict): conveted dictionary

    """
    output_dict = {}
    for field in mapper_list:
        if input_dict.get(field.field_name_1) is not None:
            output_dict[field.field_name_2] = field.conv(input_dict[field.field_name_1])
    return output_dict


def resolve_sex(scout_sex):
    """ Convert scout sex value to mutacc valid value"""
    if scout_sex == "1":
        mutacc_sex = "male"
    elif scout_sex == "2":
        mutacc_sex = "female"
    else:
        mutacc_sex = "unknown"
    return mutacc_sex


def resolve_parent(scout_parent):
    """ Convert parent (father/mother) value to mutacc """
    if scout_parent == "":
        mutacc_parent = "0"
    else:
        mutacc_parent = scout_parent
    return mutacc_parent


def resolve_phenotype(scout_phenotype):
    """ Convert scout phenotype to mutacc phenotype"""
    if scout_phenotype == 1:
        mutacc_phenotype = "unaffected"
    if scout_phenotype == 2:
        mutacc_phenotype = "affected"
    return mutacc_phenotype


def get_gene_string(genes):
    """
    Function to convert the 'genes' field in the scout variant document
    to a string format that can be read by mutacc
    """
    gene_fields = (
        "hgnc_symbol",
        "region_annotation",
        "functional_annotation",
        "sift_prediction",
        "polyphen_prediction",
    )

    gene_annotation_info = []
    for gene in genes:
        gene_info = "|".join([gene[ann_id] if gene.get(ann_id) else "" for ann_id in gene_fields])
        gene_annotation_info.append(gene_info)
    gene_annotation_info = ",".join(gene_annotation_info)

    return gene_annotation_info


SCOUT_TO_MUTACC_SAMPLE = (
    MAPPER("individual_id", "sample_id", str),
    MAPPER("sex", "sex", resolve_sex),
    MAPPER("phenotype", "phenotype", resolve_phenotype),
    MAPPER("father", "father", resolve_parent),
    MAPPER("mother", "mother", resolve_parent),
    MAPPER("analysis_type", "analysis_type", str),
    MAPPER("bam_file", "bam_file", str),
)

SCOUT_TO_MUTACC_CASE = (
    MAPPER("_id", "case_id", str),
    MAPPER("genome_build", "genome_build", str),
    MAPPER("panels", "panels", lambda panels: [panel["panel_name"] for panel in panels]),
    MAPPER("rank_model_version", "rank_model_version", str),
    MAPPER("sv_rank_model_version", "sv_rank_model_version", str),
    MAPPER("rank_score_threshold", "rank_score_threshold", int),
    MAPPER("phenotype_terms", "phenotype_terms", list),
    MAPPER("phenotype_groups", "phenotype_groups", list),
    MAPPER("diagnosis_phenotypes", "diagnosis_phenotypes", list),
    MAPPER("diagnosis_genes", "diagnosis_genes", list),
    MAPPER(
        "individuals",
        "samples",
        lambda samples: [remap(sample, SCOUT_TO_MUTACC_SAMPLE) for sample in samples],
    ),
)

SCOUT_TO_MUTACC_FORMAT = (
    MAPPER("genotype_call", "GT", str),
    MAPPER("allele_depths", "AD", lambda AD: ",".join([str(element) for element in AD])),
    MAPPER("read_depth", "DP", int),
    MAPPER("genotype_quality", "GQ", int),
    MAPPER("sample_id", "sample_id", str),
)


SCOUT_TO_MUTACC_VARIANTS = (
    MAPPER("chromosome", "CHROM", str),
    MAPPER("position", "POS", int),
    MAPPER("dbsnp_id", "ID", str),
    MAPPER("reference", "REF", str),
    MAPPER("alternative", "ALT", str),
    MAPPER("quality", "QUAL", float),
    MAPPER("filters", "FILTER", lambda filters: ",".join([str(filter) for filter in filters])),
    MAPPER("end", "END", int),
    MAPPER("rank_score", "RankScore", int),
    MAPPER("category", "category", str),
    MAPPER("sub_category", "sub_category", str),
    MAPPER("genes", "ANN", get_gene_string),
    MAPPER(
        "samples",
        "FORMAT",
        lambda samples: [remap(sample, SCOUT_TO_MUTACC_FORMAT) for sample in samples],
    ),
)



"""Commands to start MIP rare disease DNA workflow"""

import logging

import click

from cg.apps import hk, lims, scoutapi, tb
from cg.apps.environ import environ_email
from cg.cli.workflow.get_links import get_links
from cg.cli.workflow.mip.store import store as store_cmd
from cg.cli.workflow.mip_dna.deliver import CASE_TAGS, SAMPLE_TAGS
from cg.cli.workflow.mip_dna.deliver import deliver as deliver_cmd
from cg.constants import EXIT_SUCCESS, EXIT_FAIL
from cg.exc import CgError
from cg.meta.deliver import DeliverAPI
from cg.meta.workflow.mip import AnalysisAPI
from cg.store import Store
from cg.store.utils import case_exists

LOG = logging.getLogger(__name__)
EMAIL_OPTION = click.option("-e", "--email", help="email to send errors to", type=str)
PRIORITY_OPTION = click.option(
    "-p",
    "--priority",
    default="normal",
    type=click.Choice(["low", "normal", "high"]),
)
START_WITH_PROGRAM = click.option(
    "-sw", "--start-with", help="start mip from this program.", type=str
)
OPTION_DRY = click.option(
    "-d", "--dry-run", "dry_run", is_flag=True, help="print to console instead of executing"
)


@click.group("mip-dna", invoke_without_command=True)
@EMAIL_OPTION
@PRIORITY_OPTION
@START_WITH_PROGRAM
@click.option(
    "-c",
    "--case",
    "case_id",
    help="case to prepare and start an analysis for",
    type=str,
)
@click.pass_context
def mip_dna(
    context: click.Context,
    case_id: str,
    email: str,
    priority: str,
    start_with: str,
):
    """Rare disease DNA workflow"""
    context.obj["housekeeper_api"] = hk.HousekeeperAPI(context.obj)
    context.obj["trailblazer_api"] = tb.TrailblazerAPI(context.obj)
    context.obj["scout_api"] = scoutapi.ScoutAPI(context.obj)
    context.obj["lims_api"] = lims.LimsAPI(context.obj)

    context.obj["dna_api"] = AnalysisAPI(
        db=Store(context.obj["database"]),
        hk_api=context.obj["housekeeper_api"],
        tb_api=context.obj["trailblazer_api"],
        scout_api=context.obj["scout_api"],
        lims_api=context.obj["lims_api"],
        deliver_api=DeliverAPI(
            context.obj,
            hk_api=context.obj["housekeeper_api"],
            lims_api=context.obj["lims_api"],
            case_tags=CASE_TAGS,
            sample_tags=SAMPLE_TAGS,
        ),
        script=context.obj["mip-rd-dna"]["script"],
        pipeline=context.obj["mip-rd-dna"]["pipeline"],
        conda_env=context.obj["mip-rd-dna"]["conda_env"],
        root=context.obj["mip-rd-dna"]["root"],
    )
    dna_api = context.obj["dna_api"]

    if context.invoked_subcommand is not None:
        return

    if case_id is None:
        click.echo(context.get_help())
        return

    case_obj = dna_api.db.family(case_id)
    if not case_exists(case_obj, case_id):
        LOG.error(f"Case {case_id} does not exist!")
        raise click.Abort()

    all_flowcells_ondisk = dna_api.check(case_obj)
    if not all_flowcells_ondisk:
        LOG.warning(
            f"Case {case_obj.internal_id} not ready to run!",
        )
        # Commit the updates to request flowcells
        dna_api.db.commit()
        return

    # Invoke full workflow
    context.invoke(config_case, case_id=case_id)
    context.invoke(link, case_id=case_id)
    context.invoke(panel, case_id=case_id)
    context.invoke(
        run,
        case_id=case_id,
        priority=priority,
        email=email,
        start_with=start_with,
    )


@mip_dna.command()
@click.option("-c", "--case", "case_id", help="link all samples for a case", type=str)
@click.argument("sample_id", required=False)
@click.pass_context
def link(context: click.Context, case_id: str, sample_id: str):
    """Link FASTQ files for a sample_id"""
    dna_api = context.obj["dna_api"]
    link_objs = get_links(dna_api.db, case_id, sample_id)
    for link_obj in link_objs:
        LOG.info(
            "%s: %s link FASTQ files",
            link_obj.sample.internal_id,
            link_obj.sample.data_analysis,
        )
        # This block is necessary to handle cases where data analysis is not set in ClinicalDB for old samples
        if not link_obj.sample.data_analysis:
            LOG.warning(
                f"No analysis set for {link_obj.sample.internal_id}, file will still be linked"
            )
            dna_api.link_sample(sample=link_obj.sample, case_id=link_obj.family.internal_id)

        if "mip" in link_obj.sample.data_analysis.lower():
            dna_api.link_sample(sample=link_obj.sample, case_id=link_obj.family.internal_id)


@mip_dna.command("config-case")
@click.option("-d", "--dry-run", "dry_run", is_flag=True, help="print command to console")
@click.option(
    "-p",
    "--panel-bed",
    type=str,
    help="Set this option to override fetching of panel name from LIMS",
)
@click.argument("case_id", required=False, type=str)
@click.pass_context
def config_case(context: click.Context, case_id: str, panel_bed: str, dry_run: bool = False):
    """Generate a config for the case_id"""
    dna_api = context.obj["dna_api"]
    if case_id is None:
        _suggest_cases_to_analyze(context)
        return

    case_obj = dna_api.db.family(case_id)
    if not case_exists(case_obj, case_id):
        LOG.error(f"Case {case_id} does not exist!")
        raise click.Abort()

    try:
        config_data = dna_api.pedigree_config(case_obj, panel_bed=panel_bed, pipeline="mip-dna")
    except CgError as error:
        LOG.error(error.message)
        raise click.Abort()
    if dry_run:
        print(config_data)
        return
    out_path = dna_api.write_pedigree_config(config_data)
    LOG.info(f"Config file saved to {out_path}")


@mip_dna.command()
@click.option("-d", "--dry-run", "dry_run", is_flag=True, help="print output to console")
@click.argument("case_id", required=False, type=str)
@click.pass_context
def panel(context: click.Context, case_id: str, dry_run: bool = False):
    """Write aggregated gene panel file"""

    dna_api = context.obj["dna_api"]
    case_obj = dna_api.db.family(case_id)
    bed_lines = dna_api.panel(case_obj)
    if dry_run:
        for bed_line in bed_lines:
            click.echo(bed_line)
        return
    dna_api.write_panel(case_id, bed_lines)


@mip_dna.command()
@PRIORITY_OPTION
@EMAIL_OPTION
@START_WITH_PROGRAM
@click.argument("case_id", required=False, type=str)
@click.option("-d", "--dry-run", "dry_run", is_flag=True, help="print command to console")
@click.option("--mip-dry-run", "mip_dry_run", is_flag=True, help="Run MIP in dry-run mode")
@click.option(
    "--skip-evaluation",
    "skip_evaluation",
    is_flag=True,
    help="Skip mip qccollect evaluation",
)
@click.pass_context
def run(
    context: click.Context,
    case_id: str,
    dry_run: bool = False,
    email: str = None,
    mip_dry_run: bool = False,
    priority: str = None,
    skip_evaluation: bool = False,
    start_with: str = None,
):
    """Run the analysis for a case"""
    dna_api = context.obj["dna_api"]

    if case_id is None:
        _suggest_cases_to_analyze(context)
        return
    case_obj = dna_api.db.family(case_id)
    if not case_exists(case_obj, case_id):
        LOG.error(f"Case {case_id} does not exist!")
        raise click.Abort()

    email = email or environ_email()
    kwargs = dict(
        config=context.obj["mip-rd-dna"]["mip_config"],
        case=case_id,
        priority=priority or dna_api.get_priority(case_obj),
        email=email,
        dryrun=mip_dry_run,
        start_with=start_with,
        skip_evaluation=skip_evaluation,
    )

    if dna_api.tb.is_analysis_ongoing(case_id=case_obj.internal_id):
        LOG.warning("%s: analysis is ongoing - skipping", case_obj.internal_id)
        return

    if dna_api.get_skip_evaluation_flag(case_obj=case_obj):
        kwargs["skip_evaluation"] = True

    if dry_run:
        dna_api.run_command(dry_run=dry_run, **kwargs)
        return

    dna_api.run_command(**kwargs)

    if mip_dry_run:
        LOG.info("Executed MIP in dry-run mode - skipping Trailblazer step")
        return

    dna_api.tb.mark_analyses_deleted(case_id=case_id)
    dna_api.tb.add_pending_analysis(case_id=case_id, email=email)
    dna_api.set_statusdb_action(case_id=case_id, action="running")
    LOG.info("MIP rd-dna run started!")


@mip_dna.command()
@click.option("-d", "--dry-run", "dry_run", is_flag=True, help="print command to console")
@click.pass_context
def start(context: click.Context, dry_run: bool = False):
    """Start all cases that are ready for analysis"""
    dna_api = context.obj["dna_api"]
    exit_code = EXIT_SUCCESS
    cases = [case_obj for case_obj in dna_api.db.cases_to_analyze(pipeline="mip")]
    for case_obj in cases:
        if not dna_api.is_dna_only_case(case_obj):
            LOG.warning("%s: contains non-dna samples - skipping", case_obj.internal_id)
            continue
        LOG.info("%s: start analysis", case_obj.internal_id)
        has_started = dna_api.tb.has_analysis_started(case_id=case_obj.internal_id)
        if has_started:
            status = dna_api.tb.get_analysis_status(case_id=case_obj.internal_id)
            LOG.warning("%s: analysis is %s - skipping", case_obj.internal_id, status)
            continue
        if dry_run:
            continue
        try:
            context.invoke(
                mip_dna,
                priority=dna_api.get_priority(case_obj),
                case_id=case_obj.internal_id,
            )
        except (CgError, click.Abort) as error:
            LOG.error(error.message)
            exit_code = EXIT_FAIL
        except Exception as e:
            LOG.error(f"Unspecified error occurred - {e}")
            exit_code = EXIT_FAIL
    if exit_code:
        raise click.Abort()


def _suggest_cases_to_analyze(context: click.Context, show_as_error: bool = False):
    """Suggest cases to analyze"""
    if show_as_error:
        LOG.error("provide a case, suggestions:")
    else:
        LOG.warning("provide a case, suggestions:")
    for case_obj in context.obj["dna_api"].db.cases_to_analyze(pipeline="mip", limit=50):
        LOG.info(case_obj)


mip_dna.add_command(store_cmd)
mip_dna.add_command(deliver_cmd)



import gzip
import logging
import re
from pathlib import Path
from typing import List, Any
from ruamel.yaml import safe_load
import datetime as dt

from cg.apps import tb, hk, scoutapi, lims
from cg.apps.mip.base import MipAPI
from cg.apps.mip.confighandler import ConfigHandler
from cg.meta.deliver import DeliverAPI
from cg.store import models, Store
from cg.exc import CgDataError, LimsDataError
from cg.constants import (
    COMBOS,
    COLLABORATORS,
    MASTER_LIST,
    DEFAULT_CAPTURE_KIT,
    FAMILY_ACTIONS,
)

LOG = logging.getLogger(__name__)


class AnalysisAPI(ConfigHandler, MipAPI):
    """The workflow is accessed through Trailblazer but cg provides additional conventions and
    hooks into the status database that makes managing analyses simpler"""

    def __init__(
        self,
        db: Store,
        hk_api: hk.HousekeeperAPI,
        scout_api: scoutapi.ScoutAPI,
        tb_api: tb.TrailblazerAPI,
        lims_api: lims.LimsAPI,
        deliver_api: DeliverAPI,
        script: str,
        pipeline: str,
        conda_env: str,
        root: str,
        yaml_loader=safe_load,
        path_api=Path,
        logger=logging.getLogger(__name__),
    ):
        self.db = db
        self.tb = tb_api
        self.hk = hk_api
        self.scout = scout_api
        self.lims = lims_api
        self.deliver = deliver_api
        self.yaml_loader = yaml_loader
        self.pather = path_api
        self.log = logger
        self.script = script
        self.pipeline = pipeline
        self.conda_env = conda_env
        self.root = root

    def check(self, family_obj: models.Family) -> bool:
        """Check stuff before starting the analysis."""
        flowcells = self.db.flowcells(family=family_obj)
        statuses = []
        for flowcell_obj in flowcells:
            self.log.debug("%s: checking flowcell", flowcell_obj.name)
            statuses.append(flowcell_obj.status)
            if flowcell_obj.status == "removed":
                self.log.info("%s: requesting removed flowcell", flowcell_obj.name)
                flowcell_obj.status = "requested"
            elif flowcell_obj.status != "ondisk":
                self.log.warning("%s: %s", flowcell_obj.name, flowcell_obj.status)
        return all(status == "ondisk" for status in statuses)

    @staticmethod
    def get_priority(family_obj: models.Family) -> str:
        """Fetch priority for case id"""
        if family_obj.priority == 0:
            return "low"
        if family_obj.priority > 1:
            return "high"
        return "normal"

    def pedigree_config(
        self, family_obj: models.Family, pipeline: str, panel_bed: str = None
    ) -> dict:
        """Make the MIP pedigree config. Meta data for the family is taken from the family object
        and converted to MIP format via trailblazer.

        Args:
            family_obj (models.Family):
            pipeline (str): the name of the pipeline to validate the config against
            panel_bed(str): Overrides default capture kit or bypassed getting panel BED from LIMS

        Returns:
            dict: config_data (MIP format)
        """
        data = self.build_config(family_obj, pipeline=pipeline, panel_bed=panel_bed)

        # Validate and reformat to MIP pedigree config format
        config_data = self.make_pedigree_config(data=data, pipeline=pipeline)
        return config_data

    def get_target_bed_from_lims(self, sample_id: str) -> str:
        """Get target bed filename from lims"""
        target_bed_shortname = self.lims.capture_kit(sample_id)
        if not target_bed_shortname:
            raise LimsDataError("Target bed %s not found in LIMS" % target_bed_shortname)
        bed_version_obj = self.db.bed_version(target_bed_shortname)
        if not bed_version_obj:
            raise CgDataError("Bed-version %s does not exist" % target_bed_shortname)
        return bed_version_obj.filename

    def build_config(self, family_obj: models.Family, pipeline: str, panel_bed: str = None) -> dict:
        """Fetch data for creating a MIP pedigree config file"""

        def get_sample_data(link_obj):
            return {
                "sample_id": link_obj.sample.internal_id,
                "sample_display_name": link_obj.sample.name,
                "analysis_type": link_obj.sample.application_version.application.analysis_type,
                "sex": link_obj.sample.sex,
                "phenotype": link_obj.status,
                "expected_coverage": link_obj.sample.application_version.application.min_sequencing_depth,
            }

        def config_dna_sample(self, link_obj):
            sample_data = get_sample_data(link_obj)
            if sample_data["analysis_type"] == "wgs":
                sample_data["capture_kit"] = panel_bed or DEFAULT_CAPTURE_KIT
            else:
                sample_data["capture_kit"] = panel_bed or self.get_target_bed_from_lims(
                    link_obj.sample.internal_id
                )
            if link_obj.mother:
                sample_data["mother"] = link_obj.mother.internal_id
            if link_obj.father:
                sample_data["father"] = link_obj.father.internal_id
            return sample_data

        def config_rna_sample(self, link_obj):
            sample_data = get_sample_data(link_obj)
            if link_obj.mother:
                sample_data["mother"] = link_obj.mother.internal_id
            if link_obj.father:
                sample_data["father"] = link_obj.father.internal_id
            return sample_data

        dispatch = {
            "mip-dna": config_dna_sample,
            "mip-rna": config_rna_sample,
        }

        data = {
            "case": family_obj.internal_id,
            "default_gene_panels": family_obj.panels,
        }
        config_sample = dispatch[pipeline]
        data["samples"] = [config_sample(self, link_obj=link_obj) for link_obj in family_obj.links]
        return data

    @staticmethod
    def fastq_header(line: str) -> dict:
        """handle Illumina's two different header formats
        @see https://en.wikipedia.org/wiki/FASTQ_format

        @HWUSI-EAS100R:6:73:941:1973#0/1

            HWUSI-EAS100R   the unique instrument name
            6   flowcell lane
            73  tile number within the flowcell lane
            941     'x'-coordinate of the cluster within the tile
            1973    'y'-coordinate of the cluster within the tile
            #0  index number for a multiplexed sample (0 for no indexing)
            /1  the member of a pair, /1 or /2 (paired-end or mate-pair reads only)

        Versions of the Illumina pipeline since 1.4 appear to use #NNNNNN
        instead of #0 for the multiplex ID, where NNNNNN is the sequence of the
        multiplex tag.

        With Casava 1.8 the format of the '@' line has changed:

        @EAS139:136:FC706VJ:2:2104:15343:197393 1:Y:18:ATCACG

            EAS139  the unique instrument name
            136     the run id
            FC706VJ     the flowcell id
            2   flowcell lane
            2104    tile number within the flowcell lane
            15343   'x'-coordinate of the cluster within the tile
            197393  'y'-coordinate of the cluster within the tile
            1   the member of a pair, 1 or 2 (paired-end or mate-pair reads only)
            Y   Y if the read is filtered, N otherwise
            18  0 when none of the control bits are on, otherwise it is an even number
            ATCACG  index sequence


        TODO: add unit test
        """

        rs = {"lane": None, "flowcell": None, "readnumber": None}

        parts = line.split(":")
        if len(parts) == 5:  # @HWUSI-EAS100R:6:73:941:1973#0/1
            rs["lane"] = parts[1]
            rs["flowcell"] = "XXXXXX"
            rs["readnumber"] = parts[-1].split("/")[-1]
        if len(parts) == 10:  # @EAS139:136:FC706VJ:2:2104:15343:197393 1:Y:18:ATCACG
            rs["lane"] = parts[3]
            rs["flowcell"] = parts[2]
            rs["readnumber"] = parts[6].split(" ")[-1]
        if len(parts) == 7:  # @ST-E00201:173:HCLCGALXX:1:2106:22516:34834/1
            rs["lane"] = parts[3]
            rs["flowcell"] = parts[2]
            rs["readnumber"] = parts[-1].split("/")[-1]

        return rs

    @staticmethod
    def name_file(
        lane: int,
        flowcell: str,
        sample: str,
        read: int,
        undetermined: bool = False,
        date: dt.datetime = None,
        index: str = None,
    ) -> str:
        """Name a FASTQ file following MIP conventions."""
        flowcell = f"{flowcell}-undetermined" if undetermined else flowcell
        date_str = date.strftime("%y%m%d") if date else "171015"
        index = index if index else "XXXXXX"
        return f"{lane}_{date_str}_{flowcell}_{sample}_{index}_{read}.fastq.gz"

    def link_file(self, family: str, sample: str, analysis_type: str, files: List[dict]):
        """Link FASTQ files for a sample."""
        fastq_dir = Path(self.root) / family / analysis_type / sample / "fastq"
        fastq_dir.mkdir(parents=True, exist_ok=True)
        for fastq_data in files:
            fastq_path = Path(fastq_data["path"])
            fastq_name = self.name_file(
                lane=fastq_data["lane"],
                flowcell=fastq_data["flowcell"],
                sample=sample,
                read=fastq_data["read"],
                undetermined=fastq_data["undetermined"],
            )
            dest_path = fastq_dir / fastq_name
            if not dest_path.exists():
                LOG.info(f"linking: {fastq_path} -> {dest_path}")
                dest_path.symlink_to(fastq_path)
            else:
                LOG.debug(f"destination path already exists: {dest_path}")

    def link_sample(self, sample: models.Sample, case_id: str):
        """Link FASTQ files for a sample."""
        file_objs = self.hk.files(bundle=sample.internal_id, tags=["fastq"])
        files = []

        for file_obj in file_objs:
            # figure out flowcell name from header
            with gzip.open(file_obj.full_path) as handle:
                header_line = handle.readline().decode()
                header_info = self.fastq_header(header_line)

            data = {
                "path": file_obj.full_path,
                "lane": int(header_info["lane"]),
                "flowcell": header_info["flowcell"],
                "read": int(header_info["readnumber"]),
                "undetermined": ("_Undetermined_" in file_obj.path),
            }
            # look for tile identifier (HiSeq X runs)
            matches = re.findall(r"-l[1-9]t([1-9]{2})_", file_obj.path)
            if len(matches) > 0:
                data["flowcell"] = f"{data['flowcell']}-{matches[0]}"
            files.append(data)

        self.link_file(
            family=case_id,
            sample=sample.internal_id,
            analysis_type=sample.application_version.application.analysis_type,
            files=files,
        )

    def panel(self, family_obj: models.Family) -> List[str]:
        """Create the aggregated panel file."""
        all_panels = self.convert_panels(family_obj.customer.internal_id, family_obj.panels)
        bed_lines = self.scout.export_panels(all_panels)
        return bed_lines

    def write_panel(self, case_id: str, content: List[str]):
        """Write the gene panel to case dir"""
        out_dir = Path(self.root, case_id)
        out_dir.mkdir(parents=True, exist_ok=True)
        out_path = Path(out_dir, "gene_panels.bed")
        with out_path.open("w") as out_handle:
            out_handle.write("\n".join(content))

    @staticmethod
    def convert_panels(customer: str, default_panels: List[str]) -> List[str]:
        """Convert between default panels and all panels included in gene list."""
        # check if all default panels are part of master list
        if customer in COLLABORATORS and all(panel in MASTER_LIST for panel in default_panels):
            return MASTER_LIST

        # the rest are handled the same way
        all_panels = set(default_panels)

        # fill in extra panels if selection is part of a combo
        for panel in default_panels:
            if panel in COMBOS:
                for extra_panel in COMBOS[panel]:
                    all_panels.add(extra_panel)

        # add OMIM to every panel choice
        all_panels.add("OMIM-AUTO")

        return list(all_panels)

    def _get_latest_raw_file(self, family_id: str, tag: str) -> Any:
        """Get a python object file for a tag and a family ."""

        analysis_files = self.deliver.get_post_analysis_files(
            case=family_id, version=False, tags=[tag]
        )
        if analysis_files:
            analysis_file_raw = self._open_bundle_file(analysis_files[0].path)
        else:
            raise self.log.warning(
                "No post analysis files received from DeliverAPI for '%s'",
                family_id,
            )

        return analysis_file_raw

    def _open_bundle_file(self, relative_file_path: str) -> Any:
        """Open a bundle file and return it as an Python object."""

        full_file_path = self.pather(self.deliver.get_post_analysis_files_root_dir()).joinpath(
            relative_file_path
        )
        open_file = self.yaml_loader(self.pather(full_file_path).open())
        return open_file

    def get_latest_metadata(self, family_id: str) -> dict:
        """Get the latest trending data for a family."""

        mip_config_raw = self._get_latest_raw_file(family_id=family_id, tag="mip-config")
        qcmetrics_raw = self._get_latest_raw_file(family_id=family_id, tag="qcmetrics")
        sampleinfo_raw = self._get_latest_raw_file(family_id=family_id, tag="sampleinfo")
        trending = dict()
        if mip_config_raw and qcmetrics_raw and sampleinfo_raw:
            try:
                trending = self.tb.get_trending(
                    mip_config_raw=mip_config_raw,
                    qcmetrics_raw=qcmetrics_raw,
                    sampleinfo_raw=sampleinfo_raw,
                )
            except KeyError as error:
                self.log.warning(
                    "get_latest_metadata failed for '%s', missing key: %s",
                    family_id,
                    error.args[0],
                )
                trending = dict()
        return trending

    @staticmethod
    def is_dna_only_case(case_obj: models.Family) -> bool:
        """Returns True if all samples of a case has dna application type"""

        for _link in case_obj.links:
            if _link.sample.application_version.application.analysis_type in "wts":
                return False
        return True

    def get_skip_evaluation_flag(self, case_obj: models.Family) -> bool:
        """If any sample in this case is downsampled or external, returns true"""
        for link_obj in case_obj.links:
            downsampled = isinstance(link_obj.sample.downsampled_to, int)
            external = link_obj.sample.application_version.application.is_external
            if downsampled or external:
                self.log.info(
                    "%s: downsampled/external - skip evaluation",
                    link_obj.sample.internal_id,
                )
                return True
        return False

    def set_statusdb_action(self, case_id: str, action: str) -> None:
        if action in [None, *FAMILY_ACTIONS]:
            case_object = self.db.family(case_id)
            case_object.action = action
            self.db.commit()
            return
        LOG.warning(
            f"Action '{action}' not permitted by StatusDB and will not be set for case {case_id}"
        )
